"""NOTE: This module is a modified version of the ImageViewer class from the textual_imageview package.
See LICENSES file for more information."""

import math

from PIL import Image
from textual import events
from textual.app import RenderResult
from textual.widget import Widget

from textual_imageview.img import ImageView

class ImageViewer(Widget):
    """NOTE: This class is a modified version of the ImageViewer class from the textual_imageview package."""


    DEFAULT_CSS = """
    ImageViewer{
        min-width: 8;
        min-height: 8;
    }
    """

    def __init__(
            self,
            image: Image.Image,
            min_zoom: int = 10,
            max_zoom: int = 1,
            nested: bool = False,
            id: str | None = None,
        ):
        """A widget that displays an image and allows zooming and panning.
        
        Args:
            image (Image.Image): The image to display.
            min_zoom (int, optional): The minimum zoom level. Defaults to 10.
            max_zoom (int, optional): The maximum zoom level. Defaults to 1.
            nested (bool, optional): Whether the ImageViewer will be a child inside another Widget. Defaults to False.
            id: The ID of the widget in the DOM.

        Setting `nested` to True will make the ImageViewer only capture the mouse scroll up/down when its parent has focus.
        This is useful when it's a child of another Widget, or you just want to control when it's allowed to zoom.
        
        Note: passing through *args and **kwargs to Widget class is blocked except for the `id` parameter."""
        
        super().__init__(id=id)
        if not isinstance(image, Image.Image):
            raise TypeError(
                f"Expected PIL Image, but received '{type(image).__name__}' instead."
            )

        self.image = ImageView(image)
        self.mouse_down = False
        self.min_zoom = min_zoom
        self.max_zoom = max_zoom
        self.nested = nested

    def on_show(self):
        w, h = self.size.width, self.size.height
        img_w, img_h = self.image.size

        # Compute zoom such that image fits in container
        zoom_w = math.log(max(w, 1) / img_w, self.image.ZOOM_RATE)
        zoom_h = math.log((max(h, 1) * 2) / img_h, self.image.ZOOM_RATE)
        zoom = max(0, math.ceil(max(zoom_w, zoom_h)))
        self.image.set_zoom(zoom)

        # Position image in center of container
        img_w, img_h = self.image.zoomed_size
        self.image.origin_position = (-round((w - img_w) / 2), -round(h - img_h / 2))
        self.image.set_container_size(w, h, maintain_center=False)

        self.refresh()

    def on_mouse_scroll_down(self, event: events.MouseScrollDown):
        """scroll down to zoom out"""

        if not self.nested:

            offset = self.region.offset
            zoom_position = self.image.rowcol_to_xy(event.y, event.x, (offset.y, offset.x)) 
            zoom_level = self.image._zoom

            if zoom_level < self.min_zoom:
                self.image.zoom(1, zoom_position)

            self.log.event(f"Zoom level: {zoom_level}")

            self.refresh()
            event.stop()

    def on_mouse_scroll_up(self, event: events.MouseScrollDown):
        """scroll up to zoom in"""

        if not self.nested:

            offset = self.region.offset
            zoom_position = self.image.rowcol_to_xy(event.y, event.x, (offset.y, offset.x))
            zoom_level = self.image._zoom

            if zoom_level > self.max_zoom:
                self.image.zoom(-1, zoom_position)

            self.log.event(f"Zoom level: {zoom_level}")

            self.refresh()
            event.stop()

    def on_mouse_down(self, _: events.MouseDown):
        if not self.nested:
            self.mouse_down = True
            self.capture_mouse(capture=True)

    def on_mouse_up(self, _: events.MouseDown):
        if not self.nested:
            self.mouse_down = False
            self.capture_mouse(capture=False)

    def on_mouse_move(self, event: events.MouseMove):
        if self.mouse_down and (event.delta_x != 0 or event.delta_y != 0):
            self.image.move(event.delta_x, event.delta_y * 2)
            self.refresh()

    def on_resize(self, event: events.Resize):
        self.image.set_container_size(event.size.width, event.size.height)
        self.refresh()

    def render(self) -> RenderResult:
        return self.image


# Standard library imports
from __future__ import annotations
from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from textual.app import ComposeResult 
    from textual.timer import Timer
    # from textual.events import Click

# Third party imports
from rich.emoji import Emoji

# Textual imports
from textual import on
from textual.containers import Horizontal
from textual.widgets import Label, Static

# Textualdon imports
from textualdon.simplebutton import SimpleButton
from textualdon.messages import SwitchMainContent

class TopBarType(Horizontal):

    def on_mount(self):
        for child in self.children:
            child.can_focus = False

class TopBarMenu(TopBarType):

    def compose(self) -> ComposeResult:
        yield SimpleButton(f"{Emoji('house')}",    id="home",             name="Home",          classes="topbar_button")   
        yield SimpleButton(f"{Emoji('bell')}",     id="notifications",    name="Notifications", classes="topbar_button")  
        yield SimpleButton("#",                    id="explore",          name="Explore",       classes="topbar_button")     
        yield SimpleButton(f"{Emoji('earth_americas')}", id="live_feeds", name="Live Feeds",    classes="topbar_button") 
        yield SimpleButton("@",                 id="private_mentions", name="Private Mentions", classes="topbar_button")     
        yield SimpleButton(f"{Emoji('bookmark')}", id="bookmarks",        name="Bookmarks",     classes="topbar_button")      
        yield SimpleButton(f"{Emoji('star2')}",    id="favorites",        name="Favorites",     classes="topbar_button")      
        yield SimpleButton(f"{Emoji('scroll')}",   id="lists",            name="Lists",         classes="topbar_button")

class StatusWidget(TopBarType):

    def compose(self) -> ComposeResult:
        yield Static("Status: Offline", id="online_status")
        yield SimpleButton("≡", id="login_page", name="Settings", classes="topbar_button status")

class TopBar(TopBarType):

    def compose(self) -> ComposeResult:
        yield SimpleButton(f"{Emoji('arrow_backward')}", id="back", name="Back", classes="topbar_button back")
        yield TopBarMenu()
        yield Label(id="topbar_label")
        yield StatusWidget()

    def on_mount(self):
        self.online_status = self.query_one("#online_status")
        self.topbar_label = self.query_one("#topbar_label")

    def update(self, status: str, instance_url: str = None) -> None:
        self.online_status.update(status)
        self.online_status.tooltip = instance_url

    @on(SimpleButton.Pressed)
    def change_page(self, event: SimpleButton.Pressed) -> None:
        self.post_message(SwitchMainContent(event.button.id))

    @on(SimpleButton.HoverEnter)
    def show_label(self, event: SimpleButton.HoverEnter) -> None:
        self.topbar_label.update(event.button.name)

    @on(SimpleButton.HoverLeave)
    def hide_label(self, event: SimpleButton.HoverLeave) -> None:
        self.topbar_label.update("")

    

class BottomBar(Horizontal):

    class BindingLabel(Horizontal):
        def __init__(self, binding, label, **kwargs):
            super().__init__(**kwargs)
            self.binding = binding
            self.label = label

        def compose(self) -> ComposeResult:
            yield Static(self.binding, classes="bottombar_label one")
            yield Static(self.label, classes="bottombar_label two")

    def compose(self) -> ComposeResult:
        yield self.BindingLabel("Tab | Shift+Tab", "Focus")
        yield self.BindingLabel("F1-F8", "Page")
        yield self.BindingLabel("F9", "Settings")
        yield self.BindingLabel("^r", "Refresh Page")
        yield self.BindingLabel("F12", "Back", id="hide_bottombar")        

class MessageBarWidget(Horizontal):
    """This is a simple widget that displays a message at the top bar.
    It's used across the program to relay simple messages.
    It's controlled by the UpdateBannerMessage message which is in the messages.py file."""

    # notification: Reactive[str] = Reactive("")  # Store the current notification
    timer: Timer = None  # Timer instance
    clear_time = 5       # seconds to clear the message

    def compose(self) -> ComposeResult:
        yield Label(id="message_widget_text")
        Label.update

    def on_mount(self):
        self.message_widget = cast(Label, self.query_one("#message_widget_text"))


    async def update(self, message: str) -> None:
        """This overrides the normal update to add the timer.
        Activated by update_message in __init__.py"""

        self.message_widget.update(message)
        if self.timer:
            self.timer.stop()
        self.timer = self.set_timer(self.clear_time, self.clear_message)

    def clear_message(self) -> None:
        self.message_widget.update("")
        self.timer = None  


class SafeModeBar(Horizontal):

    def compose(self) -> ComposeResult:
        yield SimpleButton(
            "Safe Mode - Click here or press ctrl-d to disable",
            id="safe_mode", name="Safe Mode", classes="topbar_button safe_mode"
        )

    @on(SimpleButton.Pressed)
    def safemode_bar_disable(self) -> None:
        self.app.disable_safe_mode()

#######################################################

        

# class LeftBar(Vertical):

#     def compose(self) -> ComposeResult:
#         yield Static("", id='leftbar_content')

#     def on_mount(self):
#         # user_db = self.app.tinydb.all()
#         leftbar_content = self.query_one("#leftbar_content")
#         for item in self.app.tinydb:
#             username = item['username']
#             instance_url = item['instance_url']
#             leftbar_content.update(f"{username}\n{instance_url}")

#     def hide_if_too_small(self, width):
#         if width < 120:  # For example, hide if the width is less than 80
#             self.display = False  # This hides the widget
#         else:
#             self.display = True   # This shows the widget

# class RightBar(Vertical):

#     def compose(self) -> ComposeResult:
#         yield Static("")

#     def hide_if_too_small(self, width):
#         if width < 120:  # For example, hide if the width is less than 80
#             self.display = False  # This hides the widget
#         else:
#             self.display = True   # This shows the widget



# Standard library imports
from __future__ import annotations
from typing import TYPE_CHECKING, cast
import webbrowser
import platform
import configparser
import os
from collections import deque
from contextlib import contextmanager
from pathlib import Path
import random
import sys
from sqlite3 import DatabaseError       # for testing

if TYPE_CHECKING:
    from textual.app import ComposeResult 
    from mastodon import Mastodon

# third party imports
from mastodon import MastodonError      # for testing
import pyperclip
import clipman
from clipman.exceptions import ClipmanBaseException
from platformdirs import user_data_dir
from textual_pyfiglet.pyfiglet import figlet_format

# Textual imports
from textual import work
from textual.message import Message
# from textual.reactive import reactive
from textual.errors import TextualError     # for testing
from textual.worker import Worker, WorkerState #, WorkerFailed
from textual.events import Resize
from textual.binding import Binding
from textual.app import App, on
from textual.containers import VerticalScroll
from textual.widgets import (
    Footer,
    ContentSwitcher,
    Label,
)

# TextualDon imports
import textualdon.pages as pages
from textualdon.error_handler import ErrorHandler
from textualdon.proxy import MastodonProxy
from textualdon.bars import TopBar, MessageBarWidget, BottomBar, SafeModeBar
from textualdon.tootbox import TootBox
from textualdon.settings import Settings
from textualdon.sql import SQLite
from textualdon.messages import (
    UpdateBannerMessage,
    SuperNotify,
    LoginComplete,
    LoginStatus,
    ExamineToot,
    RefreshCurrentPage,
    SwitchMainContent,
    UserPopupMessage,
    CallbackSuccess,
    CallbackCancel,
    ScrollToWidget,
    EnableSafeMode,
    TriggerRandomError,
    ExceptionMessage,
    DeleteLogs,
    OpenRoadmap
)
from textualdon.screens import (
    WSLWarning,
    FirstWarning,
    LinkScreen,
    RoadmapScreen
)

# Rich imports and setup
# from rich.emoji import Emoji
from rich.text import Text
from rich import traceback
traceback.install()                 # setup functions must go after imports or ruff complains
        

class TextualDon(App):

    class CheckResult(Message):
        pass

    CSS_PATH = [
        "css/main.tcss",
        "css/pages.tcss",
        "css/widgets.tcss",
        "css/bars.tcss",
        "css/toot.tcss",
        "css/settings.tcss",
        "css/screens.tcss",
    ]

    BINDINGS = [
        Binding(key="f1", action="home",      description="Home",             key_display="F1", show=False),
        Binding(key="f2", action="notif",     description="Notifications",    key_display="F2", show=False),
        Binding(key="f3", action="explore",   description="Explore",          key_display="F3", show=False),
        Binding(key="f4", action="live",      description="Live Feeds",       key_display="F4", show=False),
        Binding(key="f5", action="mentions",  description="Private Mentions", key_display="F5", show=False),
        Binding(key="f6", action="bookmarks", description="Bookmarks",        key_display="F6", show=False),
        Binding(key="f7", action="favorites", description="Favorites",        key_display="F7", show=False),
        Binding(key="f8", action="lists",     description="Lists",            key_display="F8", show=False),
        Binding(key="f9", action="settings",  description="Settings",         key_display="F10", show=False),
        Binding(key="f12", action="previous_page", description="Back to Previous Page", key_display="F12", show=False),
        Binding(key="ctrl+r", action="refresh_page", description="Refresh Page", show=False),
        Binding(key="ctrl+d", action="disable_safe_mode", description="Disable Safe Mode", show=False, priority=True),
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        self.error_handler: ErrorHandler | None = None  # Error handler instance
        self.mastodon:     Mastodon | None = None       # API Wrapper instance
        self.sqlite:         SQLite | None = None       # SQLite database instance
        self.instance_url:      str | None = None       # current connected instance
        self.stored_page:       str | None = None       # stored page when the window is too small
        self.logged_in_user_id: int | None = None       # ID of the logged in user
        self.previous_pages = deque(maxlen=5)           # previous pages before the current one
        self.WSL  = False                               # Windows Subsystem for Linux flag
        self.safe_mode = False                          # Safe mode flag
        self.error = False                              # Flag for error state
        self.init_complete = False

        self.current_os = platform.system()
        self.release = platform.uname().release.lower() # nwrite comment: 

        if 'linux' in self.current_os.lower() and 'microsoft' in self.release.lower():
            self.WSL = True

        self.app_name = "textualdon"
        self.data_dir: Path = Path(user_data_dir(appname=self.app_name, ensure_exists=True))    # using platformdirs library

        base_dir = os.path.dirname(os.path.abspath(__file__))
        config_path = os.path.join(base_dir, "config.ini")
        self.config = configparser.ConfigParser()      #~ Access the configs globally with self.app.config

        cfg_read = self.config.read(config_path)
        if not cfg_read:
            raise FileNotFoundError("config.ini file not found.")
        
        ##~ Development settings ~##
        self.delete_db_on_start = self.config.getboolean("MAIN", "delete_db_on_start")
        self.force_no_clipman   = self.config.getboolean("MAIN", "force_no_clipman")
        self.force_no_pyperclip = self.config.getboolean("MAIN", "force_no_pyperclip")
        self.text_insert_time = self.config.getfloat("MAIN", "text_insert_time") #* used in many places

        # These are premade figlets for prettying up the dev console.
        # Because I'm just super fancy like that.
        self.breaker_figlet = figlet_format("-----------", font="smblock").strip() + "\n"  
        self.logo_figlet = figlet_format("  TextualDon  ", font="smblock")

    def compose(self) -> ComposeResult:
        """Create child widgets for the app."""

        self.log(Text(f"{self.breaker_figlet}\n{self.logo_figlet}", style="cyan"))

        #~ 3 classes in the app are Hidden classes that inherit from DOMNode:
        # ErrorHandler   (app.error_handler)  |  error_handler.py
        # SQLite         (app.sqlite)         |  sql.py
        # MastodonProxy  (app.mastodon)       |  proxy.py

        try:
            self.error_handler = ErrorHandler(self.data_dir)
        except Exception as e:
            raise e         # if this doesnt work, crash the app.

        with self.capture_exceptions():     # access the SQLite database globally with self.app.sqlite
            self.sqlite = SQLite(
                app_name=self.app_name,
                data_dir=self.data_dir,
                sql_script="create_tables.sql",
                del_on_start=self.delete_db_on_start
            )
        if self.error:
            return
        else:
            self.initial_page = self.sqlite.fetchone(
                "SELECT value FROM settings WHERE name= ?", ("show_on_startup",)
            )[0]
            self.previous_page = self.initial_page

            hatching = self.sqlite.fetchone(
                "SELECT value FROM settings WHERE name= ?", ("hatching",)
            )[0]
            hatching = None if hatching == 'none' else hatching

            self.theme = self.sqlite.fetchone(
                "SELECT value FROM settings WHERE name= ?", ("current_theme",)
            )[0]

        if self.force_no_clipman:
            self.clipman_works = False
        else:
            try:
                self.clipman_works = self.test_clipman()
            except:  # noqa
                self.clipman_works = False
        if self.force_no_pyperclip:
            self.pyperclip_works = False
        else:
            try:
                self.pyperclip_works = self.test_pyperclip()
            except: # noqa
                self.pyperclip_works = False

        self.log(
            f"Current OS: {self.current_os}\n"
            f"Release: {self.release}\n"
            f"Python version: {platform.python_version()}\n"
            f"WSL: {self.WSL}\n"
            f"Clipman available: {self.clipman_works}\n"
            f"Pyperclip available: {self.pyperclip_works}\n"
            f"Database location: \n{self.sqlite.user_db_path}\n"
        )

        if self.delete_db_on_start:
            self.log.warning(Text("You are in Development Mode. Database is deleted on start.", style="red"))

        with self.capture_exceptions():
            yield TopBar(id="topbar", classes="topbar")
            yield SafeModeBar(id="safe_mode_bar", classes="topbar safemode")
            yield MessageBarWidget(id="message_widget", classes="topbar message")
            yield TootBox(id="main_tootbox")
            with VerticalScroll(id="main_scroll", classes=f"main_scroll {hatching}"):
                with ContentSwitcher(initial=self.initial_page, id="main_switcher"):   
                    yield pages.LoginPage(    id="login_page",  classes=f"page_area {hatching}")                  
                    yield pages.HomePage(     id="home",        classes=f"page_area {hatching}")                        
                    yield pages.NotificationsPage(id="notifications", classes=f"page_area {hatching}")     
                    yield pages.ExplorePage(  id="explore",     classes=f"page_area {hatching}")                    
                    yield pages.LiveFeeds(    id="live_feeds",  classes=f"page_area {hatching}")                 
                    yield pages.PrivateMentionsPage(id="private_mentions", classes=f"page_area {hatching}") 
                    yield pages.BookmarksPage(id="bookmarks",   classes=f"page_area {hatching}")  
                    yield pages.FavoritesPage(id="favorites",   classes=f"page_area {hatching}")  
                    yield pages.ListsPage(    id="lists",       classes=f"page_area {hatching}")            
                    yield pages.TootPage(     id="toot_page",   classes=f"page_area {hatching}")
                    yield pages.UserProfilePage(id="user_page", classes=f"page_area {hatching}")
                    yield pages.TooSmallPage( id="too_small",   classes=f"page_area {hatching}")
                    yield pages.AboutPage(    id="about",       classes=f"page_area {hatching}")
                    yield pages.DevelopmentPage(id="dev_settings", classes=f"page_area {hatching}")
            yield BottomBar(id="bottombar")
            yield Footer(Label("TextualDon"))

    async def on_mount(self):

        if self.error:
            return

        # set convenience properties
        self.main_switcher = cast(ContentSwitcher, self.query_one("#main_switcher"))
        self.main_scroll   = cast(VerticalScroll, self.query_one("#main_scroll"))
        self.main_tootbox  = self.query_one("#main_tootbox")
        self.topbar        = self.query_one("#topbar")
        self.bottom_bar    = self.query_one("#bottombar")
        self.safemode_bar  = self.query_one("#safe_mode_bar")
        self.messagebar    = self.query_one("#message_widget")
        self.login_status  = self.query_one("#login_status")
        self.oauth_widget  = self.query_one("#oauth_widget")

        self.safemode_bar.display = False

        login_page = self.main_switcher.query_one("#login_page")
        self.settings_widget = login_page.query_one("#settings_widget")

        self.main_scroll.can_focus = False 

        self.init_complete = True
        self.log(Text(f"Initialization complete. \n {self.breaker_figlet}", style="green"))

        row1 = self.sqlite.fetchone("SELECT value FROM settings WHERE name= ?", ("show_images",))
        row2 = self.sqlite.fetchone("SELECT value FROM settings WHERE name= ?", ("auto_load",))
        row3 = self.sqlite.fetchone("SELECT value FROM settings WHERE name= ?", ("link_behavior",))
        row4 = self.sqlite.fetchone("SELECT value FROM settings WHERE name= ?", ("copypaste_engine",))
        self.show_images    = (row1[0] == "True")
        self.autoload_value = (row2[0] == "True")
        self.link_behavior    = int(row3[0])       # 0 = open in browser, 1 = copy to clipboard, 2 = helper popup
        self.copypaste_engine = int(row4[0])       # 0 = clipman, 1 = pyperclip, 2 = built-in

        warning_checkbox_wsl  = self.sqlite.fetchone("SELECT value FROM settings WHERE name= ?", ("warning_checkbox_wsl",))
        warning_checkbox_wsl  = (warning_checkbox_wsl[0] == "True")
        warning_checkbox_first = self.sqlite.fetchone("SELECT value FROM settings WHERE name= ?", ("warning_checkbox_first",))
        warning_checkbox_first = (warning_checkbox_first[0] == "True")

        if self.initial_page == "login_page":
            self.focus_login()

        if self.WSL and not warning_checkbox_wsl:
            await self.push_screen(WSLWarning(classes="modal_screen", id="introscreen_wsl"), self.intro_screens_callback)
        if not warning_checkbox_first:                  # pushed after = on top / seen first.
            await self.push_screen(FirstWarning(classes="modal_screen", id="introscreen_first"), self.intro_screens_callback) 

        # Finally:   
        self.oauth_widget.saved_users_manager.check_auto_login() 

    def attach_mastodon(self, mastodon: Mastodon):
        """The proxy class runs all calls to the Mastodon API in a worker thread. \n"""

        self.log("Attaching Mastodon object.")
        self.mastodon = MastodonProxy(mastodon)

    @on(CheckResult)
    def check_result(self):
        """Sends message if API call is hanging"""

        def internal_check():
            if self.api_startflag and not self.api_finishflag:
                workers = list(self.workers._workers)
                self.log.debug(f"HIT 2 SECONDS. ACTIVE WORKERS: {workers}")
                self.post_message(SuperNotify("This is taking a long time..."))
            
        self.set_timer(2, internal_check)   #~ 2 seconds for notification.

    @work(thread=True, exit_on_error=False, group="api_call", exclusive=False)    
    def run_api_call(self, attr, *args, **kwargs):
        """This is used by the Mastodon proxy class in proxy.py
        Would have preferred to put it in that class but it did not work for strange reasons."""

        print(f"Running API call: {attr} with args: {args}, kwargs: {kwargs}")

        self.post_message(self.CheckResult())   # used above ^
        self.api_startflag = True
        self.api_finishflag = False
        
        try:
            api_result = attr(*args, **kwargs)
        except Exception as e:
            raise e
        else:
            return api_result
        finally:
            self.api_finishflag = True

    @contextmanager
    def capture_exceptions(self):
        """Context manager that captures and forwards exceptions to the async error handler. \n\n
        This is here to allow the custom ErrorHandler class to be async. Regular sync functions
        or exceptions can use this context manager to capture exceptions (replacing the `try/except` block).
        ```
        with self.capture_exceptions():
            # dangerous code here
        ``` """
        self.error = False
        try:
            yield
        except Exception as e:
            self.error = True
            tb = sys.exc_info()[2]     
            if tb is not None:
                e.__traceback__ = tb
            self.post_message(ExceptionMessage(e))

    @on(ExceptionMessage)
    async def handle_exception(self, event: ExceptionMessage):
        """Exceptions caught with capture_exceptions (above) are sent here for transfer to
        async error handler class."""
        try:
            await self.error_handler.handle_exception(event.exception)
        except Exception as e:
            self.log.error(f"Error handler failed to handle exception: {e}")
            self._handle_exception(e)

    def intro_screens_callback(self, _=None):
        """ Function to manage the two intro screens. If both are active, the first callback will \
        focus the second screen instead of the login widget. \n
        The first argument is required to be here because self.dismiss in the screen class \
        always sends an argument. If nothing is provided it sends `None`. """

        self.log(f"intro_screens_callback: Current active screen: {self.screen.id}")

        if self.screen.id in ["introscreen_wsl", "introscreen_first"]:
            self.screen.query_one("#warning_checkbox").focus()
        else:
            self.focus_login()

    def focus_login(self):
        """Triggered by intro_screens_callback above, or on_mount if there's no intro popup."""

        self.set_focus(self.oauth_widget.login_input)
        self.post_message(ScrollToWidget(self.oauth_widget.login_input))

    def on_resize(self, event: Resize):

        if self.init_complete:
            width, height = event.size
            if self.main_switcher.current != "too_small" and width < 65:
                self.stored_page = self.main_switcher.current
                self.post_message(SwitchMainContent("too_small"))
                self.main_tootbox.visible = False
                self.log.debug(f"stored_page: {self.stored_page}")

            if self.main_switcher.current == "too_small" and width > 65:
                self.post_message(SwitchMainContent(self.stored_page))
                self.main_tootbox.visible = True

    def handle_link(self, link: str):

        if self.link_behavior == 0:
            self.open_browser(link)
        elif self.link_behavior == 1:
            self.copy_to_clipboard(link)
        elif self.link_behavior == 2:
            self.push_screen(LinkScreen(link, classes="modal_screen"))

    def test_clipman(self) -> bool:

        clipman.init()
        # save original clipboard contents. Don't want to piss off the user ;)
        current_clipboard = clipman.paste()

        clipman.copy("cuExOGZorS")
        foo = clipman.paste()

        if foo == "cuExOGZorS":
            self.log.debug("Clipman test successful.")
            testpass = True
        else:
            self.log.error("Clipman test failed.")
            testpass = False

        clipman.copy(current_clipboard)
        return testpass
    
    def test_pyperclip(self):

        # save original clipboard contents. Don't want to piss off the user ;)
        current_clipboard = pyperclip.paste()

        pyperclip.copy("4ECqwBMwTS")
        foo = pyperclip.paste()

        if foo == "4ECqwBMwTS":
            self.log.debug("Pyperclip test successful.")
            testpass = True
        else:
            self.log.error("Pyperclip test failed.")
            testpass = False

        pyperclip.copy(current_clipboard)
        return testpass

    def copy_to_clipboard(self, text: str | Path) -> None:

        with self.capture_exceptions():
            if self.copypaste_engine == 0:
                super().copy_to_clipboard(text)
                self.log.debug("Copied to clipboard with built-in method.")
            elif self.copypaste_engine == 1:
                pyperclip.copy(text)
                self.log.debug("Copied to clipboard with Pyperclip.")
            elif self.copypaste_engine == 2:
                clipman.copy(text)
                self.log.debug("Copied to clipboard with Clipman.")

        if not self.error:
            self.notify("Copied to clipboard.")

    def get_history_data(self, history: list[dict]) -> tuple[list[int], int, int]:

        counts_list = [int(day_record["accounts"]) for day_record in history]
        counts_list.reverse()
        past_2_days = sum(counts_list[-2:]) # sum of the last 2 days
        past_week = sum(counts_list)
        return counts_list, past_2_days, past_week

    @work(exclusive=True, thread=True, group='browser')
    async def open_browser(self, url):

        if url is None:
            self.log.error("No URL provided to open_browser.")
            return
        
        if isinstance(url, Path):
            # check that Path exists first
            if not url.exists():
                self.log.error(f"File does not exist: {url}")
                return
            else:
                self.log.debug(f"Confirmed file exists: {url}")
            url = url.as_uri()

        self.log.debug(f"Opening browser to: {url}")

        if self.error:  # need to check if there's already an error (we're on the reporting screen)
            try:
                result = webbrowser.open(url, new=2, autoraise=False)    # 2 = new tab
            except:  # noqa         # we're already on the error screen, so swallow any new errors
                pass
            if not result:
                self.log.error("Failed to open browser.")
                self.notify("Failed to open browser.", timeout=3)
            else:
                self.notify("Browser opened.", timeout=3)
        else:
            with self.capture_exceptions():
                result = webbrowser.open(url, new=2, autoraise=False)    # 2 = new tab
            if self.error:
                return
            self.log.debug(f"Browser open result: {result}")
            if result:
                self.notify("Browser opened.", timeout=3)
            else:
                self.notify("Failed to open browser.", timeout=3)

    ###~ EVENT HANDLERS ~###

    @on(LoginComplete)
    async def login_complete(self):
        """Event sent by OauthWidget.login_stage5 when it completes.
        Refreshes page if auto-load is enabled. """

        self.log.debug(Text("Auto-loading is enabled.", style="yellow"))

        current = self.main_switcher.current
        page_obj = self.main_switcher.query_one(f"#{current}")
        if self.autoload_value and page_obj.refresh_allowed:
            await page_obj.start_refresh_page()

    @on(ScrollToWidget)
    def scroll_to_widget(self, event: ScrollToWidget) -> None:
        """Allows scrolling the main scroll area to a specific widget on the page.
        Used by:
        - app.focus_login 
        - toot.TextAreaEdit.focus (overridden method)
        - tootbox.TextAreaReply.focus (overridden method)"""

        self.log.debug(f"Scrolling to widget: {event.widget}")
        self.main_scroll.scroll_to_center(event.widget)

    @on(UserPopupMessage)
    async def handle_msg_user_popup(self, event: UserPopupMessage):
        """ This is here because the user can either follow the user, or view their profile, \n
        and page changing has to be handled in the main app. """
        message = event.message
        account = event.account         # account dict
        relation = event.relation       # relation dict

        #! TODO Following your own account is not allowed.
        # Button should be disabled if the user is viewing their own profile.        

        if message == "follow":

            with self.capture_exceptions():
                if relation['following']:
                    await self.mastodon.account_unfollow(account["id"])
                    self.log.debug(f"Unfollowed {account['display_name']}")
                    self.post_message(SuperNotify(f"Unfollowed {account['display_name']}"))  
                else:
                    await self.mastodon.account_follow(account["id"])
                    self.log.debug(f"Followed {account['display_name']}")
                    self.post_message(SuperNotify(f"Followed {account['display_name']}"))

        elif message == "profile":

            user_page = self.main_switcher.query_one("#user_page")
            user_page.update_user(account, relation)
            self.post_message(SwitchMainContent("user_page"))
            self.post_message(UpdateBannerMessage(f"Viewing profile for {account['display_name']}"))
            await user_page.start_refresh_page()    # all pages have this method

    @on(CallbackSuccess)
    def handle_callback_success(self):

        self.log.debug(
            "Received callback success event. \n"
            f"Current active screen: {self.screen}"
        )
        self.app.pop_screen()

    @on(CallbackCancel)
    def handle_callback_cancel(self):
        self.oauth_widget.cancel_callback()

    @on(UpdateBannerMessage)
    async def update_message(self, event: UpdateBannerMessage) -> None:

        self.log.debug(f"Updating banner message: {event.message}")
        await self.messagebar.update(event.message)

    @on(SuperNotify)
    async def super_notify(self, event: SuperNotify) -> None:

        await self.messagebar.update(event.message)
        self.notify(event.message)

    @on(LoginStatus)
    def update_login_status(self, event: LoginStatus) -> None:
        """Login status in the Oauth widget."""
        self.login_status.update(event.loginpage_message)
        self.topbar.update(event.statusbar, event.instance_url)

    @on(SwitchMainContent)
    async def switch_page(self, event: SwitchMainContent) -> None:

        self.log.debug(f"Switching to page: {event.content}")
        self.log.debug(f"Previous pages: {self.previous_pages}")

        if event.content == "back":
            if len(self.previous_pages) > 0:
                self.main_switcher.current = self.previous_pages.pop()
            else:
                self.log.debug("No previous pages to go back to.")
        else:

            if event.content == self.main_switcher.current:
                self.log.debug(f"Page {event.content} is already active.")
                return
            self.previous_pages.append(self.main_switcher.current)
            self.main_switcher.current = event.content

            page_obj = self.main_switcher.query_one(f"#{event.content}")

            if self.autoload_value and page_obj.refresh_allowed:
                await page_obj.start_refresh_page()
            
        self.screen.focus_next()

    @on(ExamineToot)
    async def examine_toot(self, event: ExamineToot) -> None:

        toot_page = self.main_switcher.query_one("#toot_page")
        toot_page.main_toot_id = event.toot_id   
        self.post_message(SwitchMainContent("toot_page"))
        self.post_message(UpdateBannerMessage("Expanding single toot"))
        await toot_page.start_refresh_page()

    @on(Settings.ChangeHatching)
    def change_hatching(self, event: Settings.ChangeHatching) -> None:
        value = event.changed.value
        self.log(f"Changing hatching to {value}")

        query = self.main_switcher.query_children()
        for child in query.results():
            child.set_classes(f"page_area {value}")
        self.main_scroll.set_classes(f"main_scroll {value}")

        self.main_switcher.query_one("#login_page").refresh()
        self.main_scroll.refresh()

        # The above refreshing is to make it instantly show the new hatching.
        # It does not instantly update without the manual refreshing.

    @on(EnableSafeMode)
    def enter_safe_mode(self):

        self.log.error(Text("Entering Safe Mode.", style="white on blue"))
        self.workers.cancel_all()
        self.sqlite.readonly_mode = True
        self.safemode_bar.display = True
        self.settings_widget.disabled = True
        self.safe_mode = True

    def disable_safe_mode(self, _=None):

        self.log.error(Text("Disabling Safe Mode.", style="white on blue"))
        self.sqlite.readonly_mode = False
        self.notify("Safe Mode disabled.")
        self.safemode_bar.display = False
        self.settings_widget.disabled = False
        self.error_handler.reset_stored_errors()
        self.safe_mode = False

    @on(TriggerRandomError)
    def trigger_random_error(self):

        python_errors = [
            ValueError, DatabaseError, TextualError, MastodonError, ClipmanBaseException
        ]

        error = random.choice(python_errors)

        with self.capture_exceptions():
            raise error("Random error triggered.")

    @on(DeleteLogs)
    async def delete_logs(self):

        await self.error_handler.delete_logs()

    @on(OpenRoadmap)
    async def open_roadmap(self):

        await self.push_screen(RoadmapScreen(classes="fullscreen"))

    ###~ ACTIONS ~###

    @on(RefreshCurrentPage)     
    def action_refresh_page(self) -> None:
        """ Message is sent by:
        - `tootscreens.TootOptionsMainUser.delete_toot` """

        if self.mastodon is None:
            self.notify("You are not logged in.")
            return

        page_str = self.main_switcher.current
        page = self.main_switcher.query_one(f"#{page_str}")
        self.log(f"Refreshing current page: {page}")
        if page.refresh_allowed:
            self.set_timer(0.50, page.start_refresh_page)     # half second delay to give server time to refresh.
        else:
            self.notify("This page cannot be refreshed.")

    def action_disable_safe_mode(self):
        self.log.debug("Action: Disabling Safe Mode.")
        self.disable_safe_mode()

    def action_home(self):
        self.post_message(SwitchMainContent("home"))

    def action_notif(self):
        self.post_message(SwitchMainContent("notifications"))

    def action_explore(self):
        self.post_message(SwitchMainContent("explore"))

    def action_live(self):
        self.post_message(SwitchMainContent("live_feeds"))

    def action_mentions(self):
        self.post_message(SwitchMainContent("private_mentions"))

    def action_bookmarks(self):
        self.post_message(SwitchMainContent("bookmarks"))

    def action_favorites(self):
        self.post_message(SwitchMainContent("favorites"))

    def action_lists(self):
        self.post_message(SwitchMainContent("lists"))

    def action_settings(self):
        self.post_message(SwitchMainContent("login_page"))

    def action_previous_page(self):
        self.post_message(SwitchMainContent("back"))

    ###~ REACTIVES ~###

    def watch_theme(self, value: str) -> None:

        self.log.debug(f"Theme changed to: {value}")
        self.sqlite.update_column("settings", "value", value, "name", "current_theme")


    ###~ Worker Debug stuff ~###

    @on(Worker.StateChanged)
    def worker_state_changed(self, event: Worker.StateChanged) -> None:

        if event.state == WorkerState.SUCCESS:
            self.log(Text(f"Worker {event.worker.name} completed successfully", style="green"))
        elif event.state == WorkerState.ERROR:
            self.log.error(Text(f"Worker {event.worker.name} encountered an error", style="red"))
        elif event.state == WorkerState.CANCELLED:
            self.log(Text(f"Worker {event.worker.name} was cancelled", style="yellow"))

def run():
    """Entry point for the `textualdon` command."""
    TextualDon().run()

# for running with python -m textualdon
if __name__ == "__main__":
    TextualDon().run()


html_success_msg = """
<html>
<head>
    <style>
        body {
            font-family: Arial, sans-serif;
            text-align: center;
            background-color: #f4f4f9;
            margin: 0;
            padding: 0;
            display: flex;
            justify-content: center;
            align-items: center;
            height: 100vh;
            color: #333;
        }
        div {
            background-color: #fff;
            padding: 20px;
            border-radius: 8px;
            box-shadow: 0 4px 6px rgba(0, 0, 0, 0.1);
        }
        h1 {
            margin-top: 0;
            color: #0078d4;
        }
        p {
            margin-bottom: 0;
            font-size: 1.1em;
        }
    </style>
</head>
<body>
    <div>
        <h1>TextualDon is authorized.</h1>
        <p>You can close this window now.</p>
    </div>
</body>
</html>
"""

html_failure_msg = """
<html>
<head>
    <style>
        body {
            font-family: Arial, sans-serif;
            text-align: center;
            background-color: #f4f4f9;
            margin: 0;
            padding: 0;
            display: flex;
            justify-content: center;
            align-items: center;
            height: 100vh;
            color: #333;
        }
        div {
            background-color: #fff;
            padding: 20px;
            border-radius: 8px;
            box-shadow: 0 4px 6px rgba(0, 0, 0, 0.1);
            border: 2px solid #ff4d4d; /* Red border for emphasis */
        }
        h1 {
            margin-top: 0;
            color: #ff1a1a; /* Bright red for the error title */
        }
        p {
            margin-bottom: 0;
            font-size: 1.1em;
            color: #990000; /* Darker red for the message */
        }
    </style>
</head>
<body>
    <div>
        <h1>Authorization Error</h1>
        <p>No code received. Please try again.</p>
    </div>
</body>
</html>
"""


# Standard library imports
from __future__ import annotations
from typing import Sequence, Any
from contextlib import contextmanager
from pathlib import Path
import sqlite3
from importlib import resources

# Textual imports
from textual.dom import DOMNode


class SQLite(DOMNode):
    """A simple SQLite database wrapper for TextualDon.   
    Cannot attach child widgets (blocked)."""


    def __init__(
            self,
            app_name: str,
            data_dir: Path,
            sql_script: str,
            db_filename: str = None,
            del_on_start: bool = False,
            **kwargs):
        """Create a new SQLite database wrapper.    
        Must pass in `app_name`: the name of the package that is using this widget.   
        This same name will be used to create a directory in the user's data directory to store the database file.

        `del_on_start` will delete the existing database file and create a new one (for dev mode).

        Args:
            app_name: This must be the name of the package that is using this widget.
                This MUST be the same name as your program / main package.
            data_dir: Directory on user's computer where the folder will be created.
            sql_script: The name of the SQL script file to use for setting up the database tables and schema.   
                Path is relative to the main src directory.
            db_filename: The name the database file will use. If None, name will be `<app_name>.db`.
            del_on_start: Whether to delete the database file on start. Defaults to False.
            name: The name of the widget.
            id: The ID of the widget in the DOM.
            classes: The CSS classes for the widget.

        EXAMPLE:
        ```
        db = SQLite("my_package", "create_tables.sql")
        ```
        
        """

        super().__init__(**kwargs)

        self.log(
            "SQLite widget initialized. \n"
            f"SQLite Version: {sqlite3.sqlite_version}\n"
            f"Library Version: {sqlite3.version}\n"
        )

        self.app_name:    str = app_name
        self.data_dir:   Path = data_dir
        self.sql_script:  str = sql_script
        self.db_filename: str = db_filename or f"{app_name}.db"
        self.del_on_start:  bool = del_on_start
        self.readonly_mode: bool = False

        self.user_db_path = self.data_dir / self.db_filename
        
        if self.user_db_path.exists() and self.del_on_start:
            print(f"del_on_start: {del_on_start}  |  DELETING existing database file!")
            self.user_db_path.unlink()   

        exists = self.user_db_path.exists()

        self.connection = sqlite3.connect(self.user_db_path)    
        if not exists:
            try:
                self.initialize_db()
            except Exception as e:
                e.add_note("Error initializing database.")
                raise e
    
    def initialize_db(self):

        # here it looks up the file in the package directory.
        # This is why we must pass in the app_name, so it knows what package to look for.

        sql_file_path = Path(resources.files(self.app_name) / self.sql_script)
        print(f"SQL file location: \n{sql_file_path} \n")
        with open(sql_file_path, 'r') as f:
            script = f.read()
            self.execute_script(script)

    @contextmanager
    def _cursor(self):
        """Used internally by class"""

        cursor = self.connection.cursor()
        try:
            yield cursor
        finally:
            cursor.close()

    def execute_script(self, script: str): 
        """Execute a SQL script on the database.

        EXAMPLE:
        ```
        script = "create_table.sql"
        db.execute_script(script)
        ``` """

        with self.app.capture_exceptions():
            with self._cursor() as cursor:
                cursor.executescript(script)
            self.connection.commit()
            if self.app.error:
                self.connection.rollback()
                print("ERROR executing script on database.")
            else:
                print("Successfully executed script on database.")


    def insert_one(self, table: str, columns: list[str], values: Sequence[Any], auto_commit: bool = True):
        """Insert a single row into the database.   
        Using the auto_commit parameter, you can stage multiple inserts before committing them all at once.
        Simply let it go back to the default value of True to commit all inserts.

        EXAMPLE:
        ```    
        table = "users"
        columns = ["name", "age", "email"]
        values = ["Alice", 30, "alice@example.com"]
        db.insert_one(table, columns, values)
        ```
        Raw SQL:   
        `INSERT INTO {table} ({', '.join(columns)}) VALUES (?, ?, ?);`

        Args:
            table (str): The name of the table to insert the row into.
            columns (list[str]): The list of column names.
            values (Sequence[Any]): List or tuple of values to insert.
            auto_commit (bool, optional): Whether to commit the transaction automatically. Defaults to True.
        """

        if self.readonly_mode:
            self.app.notify("Database is in read-only mode. Can't save changes.")
            return

        placeholders = ', '.join(['?'] * len(values))
        query = f"INSERT INTO {table} ({', '.join(columns)}) VALUES ({placeholders})"

        with self.app.capture_exceptions():
            with self._cursor() as cursor:
                cursor.execute(query, values)   
            if auto_commit:
                self.connection.commit()
            if self.app.error:
                self.connection.rollback()

        self.log.debug(f"Successfully inserted a row into {table} with values {values}.")


    def delete_one(self, table_name: str, column_name: str, value: Any):
        """Delete a row from a database table.

        EXAMPLE:
        ```
        db.delete_one('employees', 'id', 3)
        ```
        Raw SQL:   
        `DELETE FROM {table_name} WHERE {column_name} = ?;` """

        if self.readonly_mode:
            self.app.notify("Database is in read-only mode. Can't save changes.")
            return
        
        sql_delete_query = f"DELETE FROM {table_name} WHERE {column_name} = ?;"

        with self.app.capture_exceptions():
            with self._cursor() as cursor:
                cursor.execute(sql_delete_query, (value,))
            self.connection.commit()
            if self.app.error:
                self.connection.rollback()

        self.log.debug(f"Successfully deleted {table_name} row where {column_name} is {value}.")

    def update_column(
            self,
            table_name: str,
            column_name: str,
            new_value: Any,
            condition_column: str,
            condition_value: Any
        ):
        """Update a column in a database table.
        
        EXAMPLE:
        ```
        db.update_column('employees', 'salary', 75000, 'id', 3)
        ``` 
        Raw SQL:   
        `UPDATE {table_name} SET {column_name} = ? WHERE {condition_column} = ?; `"""

        if self.readonly_mode:
            self.app.notify("Database is in read-only mode. Can't save changes.")
            return
        
        sql_update_query = f"UPDATE {table_name} SET {column_name} = ? WHERE {condition_column} = ?;"

        with self.app.capture_exceptions():
            with self._cursor() as cursor:
                cursor.execute(sql_update_query, (new_value, condition_value))
            self.connection.commit()
            if self.app.error:
                self.connection.rollback()

        self.log.debug(
            f"Successfully updated {table_name} column {column_name} to "
            f"{new_value} where {condition_column} is {condition_value}."
        )

    def fetchall(self, query: str, params: Sequence = None) -> list[tuple]:
        """This method runs a SQL query and retrieves all rows that match the query criteria.
        
        EXAMPLE:
        ```
        query = "SELECT * FROM users WHERE name = ?"
        params = ("Alice",)   # Note the comma, (must be sequence)
        rows = db.fetchall(query, params)
        ``` 
        Or use the shortcut:
        ```
        row = db.fetchone("SELECT * FROM users WHERE name = ?", ("Alice",))
        ```
        Args:
            query (str): The SQL query to run.
            params (Sequence, optional): The query parameters. Defaults to None.

        Returns:
            list[tuple]: A list of rows that match the query criteria.
        """

        with self.app.capture_exceptions():
            with self._cursor() as cursor:
                cursor.execute(query, params or [])
                return cursor.fetchall()
            if self.app.error:
                self.connection.rollback()     

    
    def fetchone(self, query: str, params: Sequence = None) -> tuple:
        """This method is similar to fetchall, but it only retrieves a single row
        from the database, even if multiple rows meet the query criteria.
        
        EXAMPLE:
        ```
        query = "SELECT * FROM users WHERE name = ?"
        params = ("Alice",)   # Note the comma, (must be sequence)
        row = db.fetchone(query, params)
        ``` 
        Or use the shortcut:
        ```
        row = db.fetchone("SELECT * FROM users WHERE name = ?", ("Alice",))
        ```

        Args:
            query (str): The SQL query to run.
            params (Sequence, optional): The query parameters. Defaults to None.

        Returns:
            tuple: A single row that matches the query criteria.
        """
        with self.app.capture_exceptions():
            with self._cursor() as cursor:
                cursor.execute(query, params or [])
                return cursor.fetchone()
            if self.app.error:
                    self.connection.rollback()
    
    def close(self):
        """Close the database connection."""
        self.connection.close()









# Standard Library imports
from __future__ import annotations
from typing import TYPE_CHECKING, cast, Optional
from textwrap import shorten
import datetime as dt
from enum import Enum

if TYPE_CHECKING:
    from zoneinfo import ZoneInfo

# Third Party imports
from mastodon.Mastodon import Mastodon
from tzlocal import get_localzone
import humanize
from rich.emoji import Emoji
from rich.text import Text

# Textual imports
from textual.dom import NoScreen
from textual.binding import Binding
from textual import on, work # events
from textual.message import Message
from textual.containers import Horizontal, Vertical,  Container  # VerticalScroll, Grid
from textual.widget import Widget
from textual.widgets import (
    Pretty,
    Static,
    ContentSwitcher,
    TextArea,
)

# TextualDon imports
from textualdon.simplebutton import SimpleButton
from textualdon.widgets import ImageViewerWidget
from textualdon.tootbox import TootBox
from textualdon.messages import (
    ExamineToot,
    ScrollToWidget,
    SuperNotify
)
from textualdon.bs4_parser import BS4Parser
from textualdon.screens import ConfirmationScreen
from textualdon.tootscreens import (
    TootOptionsMainUser,
    TootOptionsOtherUser,
    UserPopup,
    MuteScreen,
    BlockScreen,
)


class TootWidget(Horizontal):

    class DeleteToot(Message):
        """Used by TootOptionsScreen to trigger deleting a toot."""
        def __init__(self, redraft: bool = False) -> None:
            super().__init__()
            self.redraft = redraft

    class MuteUser(Message):
        """Used by TootOptionsScreen to trigger muting a user."""
        pass

    class BlockUser(Message):
        """Used by TootOptionsScreen to trigger blocking a user."""
        pass


    BINDINGS = [
        Binding("r", "reply", "Reply", show=True),
        Binding("b", "boost", "Boost", show=True),
        Binding("f", "favourite", "Favourite", show=True),
        Binding("m", "bookmark", "Bookmark", show=True),
        Binding("o", "options", "More Options", show=True),
        Binding("enter", "switch_to_tootpage", "Expand toot", show=True),
    ]

    def __init__(self, json: dict, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.can_focus = True       #? ~ notes/widget_focus.md ~

        self.option_handler = TootOptionHandler(self)         
        self.json: dict = json

        self.json_on       = False        # for the JSON view
        self.replybox_on   = False        # toggles reply box
        self.morebox_on    = False        # toggles options box
        self.edit_mode     = False        # toggles edit mode

        # NOTE: It won't request the relationship dict until the user clicks either
        # the 'More' button (options box) or the user name (user popup).
        # If it auto requested it for every single toot, you'd probably get rate limited.

        self.reblog: dict | None = self.json["reblog"]   

        # reblog is the nested json of boosted toot; None if not boosted
        # So if reblogged/boosted, set the main json to the nested reblog json,
        # after extracting the booster's account info.

        if self.reblog:
            self.boosted_by = f"{self.json['account']['display_name']} boosted"
            self.boosted_by_id = self.json["account"]["id"]
            self.boosted_by_account_dict = self.json["account"]
            self.json = self.reblog         #* replace json with nested json
        else:
            self.boosted_by = None

        self.in_reply_to_id:         int | None = self.json["in_reply_to_id"]
        self.in_reply_to_account_id: int | None = self.json["in_reply_to_account_id"]
        self.media_attachments: list = self.json["media_attachments"]
        self.card: dict | None = self.json["card"]
        self.poll: dict | None = self.json["poll"]

    def compose(self):

        self.mastodon = cast(Mastodon, self.app.mastodon)          
        self.loading = True         # always starts with loading screen

        yield Static(id="in_reply_markerbar")
        with Container(id="toot_container", classes="content_container"):
            if self.reblog:
                yield SimpleButton(self.boosted_by, id="boosted_by_button", classes="toot_element")
            if self.in_reply_to_id:
                yield SimpleButton(id="in_reply_to_button", classes="toot_element")

            with Container(classes="toot_header"):
                yield SimpleButton(id="toot_username_button", classes="toot_element")
                yield Static(id="toot_time_label", classes="toot_element b")
                yield SimpleButton(id="toot_user_url_button", classes="toot_element")
                yield Static(id="toot_followers_label", classes="toot_element b")

            with ContentSwitcher(id="toot_content_switcher", initial="toot_content_container"):
                yield TootContentContainer(self, id="toot_content_container")  # note passing in self
                yield Pretty("", id='toot_json_container', classes="toot_extra")
                yield TootEditContainer(self, id="toot_edit_container")  # note passing in self

            with Horizontal(classes="toot_footer"):
                yield SimpleButton(id="toot_reply_button", classes="toot_element")
                yield SimpleButton(id="toot_boost_button", classes="toot_element")
                yield SimpleButton(id="toot_favourite_button", classes="toot_element")
                yield SimpleButton(id="toot_bookmark_button", classes="toot_element")
                yield SimpleButton("More", id="more", classes="toot_element")

            yield SimpleButton(id="toot_url_button", classes="toot_element")

    def on_mount(self):
        """The widget is originally composed with all the fields empty.
        This allows the program to draw the widget first and then fill in the data later."""

        self.call_after_refresh(self.load_toot_data)

    def load_toot_data(self):

        self.account = self.json["account"]

        self.toot_id: int  = self.json["id"]
        self.user_id: int  = self.account["id"]
        self.username: str = self.account["username"]
        self.display_name:   str = self.account["display_name"]
        self.toot_user_url:  str = self.account["url"]
        self.toot_url:       str = self.json["url"]
        self.toot_time: dt.datetime = self.json["created_at"]
        self.toot_followers: int = f"Followers: {self.json['account']['followers_count']}"
        self.toot_reply:     int = f"{Emoji('right_arrow_curving_left')} {self.json['replies_count']} Reply"
        self.tags:     list = self.json["tags"]
        self.emojis:   list = self.json["emojis"]     
        self.mentions: list = self.json["mentions"]
        self.visibility:  str = self.json["visibility"]
        self.sensitive:  bool = self.json["sensitive"]
        self.favourited: bool = self.json["favourited"] 
        self.reblogged:  bool = self.json["reblogged"]   
        self.muted:      bool = self.json["muted"]           
        self.bookmarked: bool = self.json["bookmarked"]
        self.relation_dict = [] # this is lazy loaded when the user popup is opened. Here as safety.

        # These are set in __init__ but we need to refresh them anyway.
        self.in_reply_to_id:         int | None = self.json["in_reply_to_id"]
        self.in_reply_to_account_id: int | None = self.json["in_reply_to_account_id"]
        self.media_attachments: list = self.json["media_attachments"]
        self.card: dict | None = self.json["card"]
        self.poll: dict | None = self.json["poll"]

        self.is_main_user = self.determine_if_main_user()

        self.toot_container = self.query_one("#toot_container")   

        self.toot_switcher          = cast(ContentSwitcher,      self.query_one("#toot_content_switcher"))   
        self.toot_content_container = cast(TootContentContainer, self.query_one("#toot_content_container"))

        in_reply_markerbar = self.query_one("#in_reply_markerbar")
        if self.in_reply_to_id:
            self.query_one("#in_reply_to_button").update(f"{Emoji('arrow_upper_left')} Continued Thread")                                  # this controls the black bar on the side that appears
        else:
            in_reply_markerbar.display = False

        self.toot_username_button = self.query_one("#toot_username_button")
        self.toot_user_url_button = self.query_one("#toot_user_url_button")
        self.toot_time_label      = self.query_one("#toot_time_label")
        self.toot_followers_label = self.query_one("#toot_followers_label")
        self.toot_reply_button    = self.query_one("#toot_reply_button")
        self.toot_favorite_button = self.query_one("#toot_favourite_button")
        self.toot_boost_button    = self.query_one("#toot_boost_button")
        self.toot_bookmark_button = self.query_one("#toot_bookmark_button")
        self.toot_url_button      = self.query_one("#toot_url_button")

        self.toot_username_button.update(f"{self.display_name} (@{self.username})")
        self.toot_user_url_button.update(f"{Emoji('link')} {self.toot_user_url}")
        self.toot_followers_label.update(self.toot_followers)
        self.toot_reply_button.update(self.toot_reply)

        self.set_time()

        self.toot_username_button.tooltip = (f"{self.display_name} (@{self.username}) \n\n"
                                    "Click to open user pop-up.")
        
        user_url_display = f"{Emoji('link')} {self.toot_user_url}"
        self.toot_user_url_button.update(user_url_display)

        toot_url_display = f"{Emoji('link')} {self.toot_url}"
        self.toot_url_button.update(toot_url_display)

        if self.favourited:
            toot_favourite = f"{Emoji('star2')} {self.json['favourites_count']} Favourited"
            toot_favourite_tooltip = "Click to unfavourite."
        else:
            toot_favourite = f"{Emoji('star')} {self.json['favourites_count']} Favourite"
            toot_favourite_tooltip = "Click to favourite."
        self.toot_favorite_button.update(toot_favourite)
        self.toot_favorite_button.tooltip = toot_favourite_tooltip

        if self.bookmarked:
            toot_bookmark = f"{Emoji('bookmark')} Bookmarked"
            toot_bookmark_tooltip = "Click to remove bookmark."
        else:
            toot_bookmark = f"{Emoji('bookmark')} Bookmark"
            toot_bookmark_tooltip = "Click to bookmark."
        self.toot_bookmark_button.update(toot_bookmark)
        self.toot_bookmark_button.tooltip = toot_bookmark_tooltip

        # Boosted/reblogged logic
        if self.reblogged:
            toot_boost = f"{Emoji('rocket')} {self.json['reblogs_count']} Boosted"
            toot_boost_tooltip = "Click to remove boost."
        else:
            toot_boost = f"{Emoji('crescent_moon')} {self.json['reblogs_count']} Boost"
            toot_boost_tooltip = "Click to boost."
        self.toot_boost_button.update(toot_boost)
        self.toot_boost_button.tooltip = toot_boost_tooltip

        self.query_one("#toot_json_container").update(self.json)

        for child in self.children:
            self.recursive_no_focus(child)

        self.loading = False

    def recursive_no_focus(self, child = None):
        """Set can_focus to False for every child in the toot widget.
        The only thing in a toot widget that's allowed to focus is the edit container,
        so we just turn that back on manually when needed."""

        if child is None:
            self.log.error("No child passed to recursive_no_focus")
            return

        child.can_focus = False
        if child.children:
            for subchild in child.children:
                self.recursive_no_focus(subchild)

    def set_time(self):

        timezone: ZoneInfo = get_localzone()    
        current_time = dt.datetime.now(timezone)
        elapsed_time: dt.timedelta = current_time - self.toot_time

        if elapsed_time.days > 3:
            time_label = humanize.naturaldate(self.toot_time)
        else:
            time_label = humanize.naturaltime(elapsed_time)

        self.toot_time_label.update(time_label)
        self.toot_time_label.tooltip = self.toot_time.strftime("%d %b %Y %H:%M")

    def on_focus(self):
        self.log.debug(f"{self.name} focused. ")
        self.styles.border = ('dashed', self.app.theme_variables["primary"])

        # TODO Need note about doing this through the code and not CSS

    def on_blur(self):
        self.styles.border = ('blank', 'transparent')

    def determine_if_main_user(self) -> bool:

        # Technically speaking, it's possible to encounter a user on a different instance with
        # the same user_id as the logged in user. This is highly unlikely, but it's possible.

        user_id = self.user_id
        if user_id == self.app.logged_in_user_id:
            return True
        else:
            return False

    def switch_content(self, switch_to: str) -> None:
        """This is used by more box (TootOptions). Switches normal/html/json mode"""

        self.toot_switcher.current = switch_to

    def edit_toot(self):
        """This is used by TootOptions. Switches to edit mode"""

        if not self.edit_mode:
            self.switch_content("toot_edit_container")
            self.query_one("#toot_edit_container").edit_toot()
            self.query_one("#toot_edit_container").focus()
            self.edit_mode = True
        else:
            self.switch_content("toot_content_container")
            self.edit_mode = False

    @work
    async def refresh_toot(self, delay: int = 0.5) -> None:
        """ This is called by:
        ```
        - tootscreens.TootOptionsScreen.refresh_toot
        - toot.TootEditContainer.save_edit             # refresh after edit
        - toot.TootOptionHandler.handle_toot_action    # refresh after actions like boost, favourite, etc.
        - toot.TootBox.post_toot                       # when Tootbox is a reply  """

        self.log.debug("Refreshing Toot.")
        self.loading = True

        async def refresh_internal():

            json = None
            with self.app.capture_exceptions():
                json = await self.mastodon.status(self.toot_id)
                if json:
                    self.json = json
                    self.call_after_refresh(self.load_toot_data)
                    self.call_after_refresh(self.toot_content_container.load_toot_content)
        self.set_timer(delay, refresh_internal)   # half second delay to give server time to refresh.


    async def view_image(self):
        await self.toot_content_container.imgviewer_widget.fullscreen()

    @on(SimpleButton.Pressed, selector='#toot_username_button')
    @work
    async def open_user_popup(self) -> None:

        # lazy load the relation_dict every time the user popup is opened
        relation_list = None
        with self.app.capture_exceptions():
            relation_list = await self.mastodon.account_relationships(self.user_id)
        if relation_list:
    
            self.log.debug(relation_list)
            self.relation_dict = relation_list[0]

            await self.app.push_screen(UserPopup(self.json['account'], self.relation_dict, classes="modal_screen"))

    @on(SimpleButton.Pressed, selector='#boosted_by_button')
    @work
    async def open_booster_popup(self) -> None:

        # lazy load the relation_dict every time the booster popup is opened
        relation_list = None
        with self.app.capture_exceptions():
            relation_list = await self.mastodon.account_relationships(self.boosted_by_id)
        if relation_list:

            relation_dict = relation_list[0]
            self.log.debug(relation_dict)

        await self.app.push_screen(UserPopup(self.boosted_by_account_dict, relation_dict, classes="modal_screen"))  

    @on(SimpleButton.Pressed, selector='#toot_user_url_button')
    def open_user_url(self) -> None:

        self.app.handle_link(self.toot_user_url)

    @on(SimpleButton.Pressed, selector='#more')
    async def show_more_options(self) -> None:

        # lazy loading the relation_dict every time the options screen is opened
        relation_list = None
        with self.app.capture_exceptions():
            relation_list = await self.mastodon.account_relationships(self.user_id)
        if relation_list:   

            self.log.debug(relation_list)
            self.relation_dict = relation_list[0]

            if self.is_main_user:
                await self.app.push_screen(TootOptionsMainUser(self, name="toot_options", classes="modal_screen"))
            else:
                await self.app.push_screen(TootOptionsOtherUser(self, name="toot_options", classes="modal_screen"))

    @on(SimpleButton.Pressed, selector='#toot_url_button')
    def open_toot_url(self) -> None:

        self.app.handle_link(self.toot_url)

    @on(SimpleButton.Pressed, selector="#toot_boost_button")
    async def boost_toot(self) -> None:
        action          = "status_unreblog" if self.reblogged else "status_reblog"
        success_message = "Boost removed"  if self.reblogged else "Toot boosted."
        failure_message = "FAILED to remove boost." if self.reblogged else "FAILED to boost toot."

        await self.option_handler.handle_toot_action(
            action=action,
            success_message=success_message,
            failure_message=failure_message,
            toggle_status=self.reblogged,
            key=self.option_handler.StatusKey.REBLOGGED
        )

    @on(SimpleButton.Pressed, selector="#toot_favourite_button")
    async def favorite_toot(self) -> None:
        action          = "status_unfavourite" if self.favourited else "status_favourite"
        success_message = "Toot Unfavourited"  if self.favourited else "Toot favourited."
        failure_message = "FAILED to Unfavourite toot." if self.favourited else "FAILED to favourite toot."

        await self.option_handler.handle_toot_action(
            action=action,
            success_message=success_message,
            failure_message=failure_message,
            toggle_status=self.favourited,
            key=self.option_handler.StatusKey.FAVOURITED,
        )

    @on(SimpleButton.Pressed, selector='#toot_bookmark_button')
    async def bookmark_toot(self) -> None:
        action         = "status_unbookmark" if self.bookmarked else "status_bookmark"
        success_message = "Toot unbookmarked." if self.bookmarked else "Toot bookmarked."
        failure_message = "FAILED to unbookmark toot." if self.bookmarked else "FAILED to bookmark toot."

        await self.option_handler.handle_toot_action(
            action=action,
            success_message=success_message,
            failure_message=failure_message,
            toggle_status=self.bookmarked,
            key=self.option_handler.StatusKey.BOOKMARKED
        )

    @on(SimpleButton.Pressed, selector='#toot_reply_button')
    async def reply_to_toot(self, event: SimpleButton.Pressed = None) -> None:

        if event:
            self.log.debug(f"Button: {event.button.id}")

        if not self.replybox_on:
            self.toot_container.mount(TootBox(toot_widget=self, id="reply_box"))
            self.replybox_on = True
        else:
            self.query_one("#reply_box").remove()
            self.replybox_on = False

    @on(SimpleButton.Pressed, selector='#in_reply_to_button')
    def switch_to_inreplyto(self) -> None:
        """This will bubble up to the main app and switch to the toot page with the toot_id"""

        self.post_message(ExamineToot(self.in_reply_to_id))


    ###~ TOOT OPTIONS CALLBACKS ~###

    @on(DeleteToot)
    async def delete_toot(self, event: DeleteToot) -> None:

        self.log.debug("Delete toot callback.")
        self.app.push_screen(
            ConfirmationScreen(classes = "modal_screen", forward=event.redraft),
            self._delete_toot
        )

    async def _delete_toot(self, redraft: bool = False) -> None:
        """ Callback from `delete_toot` (above). \n
        `redraft` is `True` if the user wants to redraft the toot after deletion."""

        await self.option_handler.handle_toot_action(
            action = "status_delete",
            success_message = "Toot deleted.",
            failure_message = "FAILED to delete toot!",
            toggle_status=False      # doesn't matter whether this is true or false. Just need to pass something.
        )
        await self.remove()

        if redraft:
            self.log.debug("Redrafting toot.")
            self.app.main_tootbox.set_text(self.toot_content_container.parsed_content)

    @on(MuteUser)
    async def mute_user(self) -> None:

        self.log.debug("Mute user callback.")

        mute_status = self.relation_dict['muting']
        if mute_status:     # dont need confirmation if unmuting
            await self._mute_user(mute_status)
        else:
            self.app.push_screen(
                MuteScreen(self.username, mute_status, classes = "modal_screen"),
                self._mute_user
            )

    async def _mute_user(self, mute_status:bool) -> None:

        action          = "account_unmute" if mute_status else "account_mute"
        success_message = f"@{self.username} unmuted." if mute_status else f"@{self.username} muted."
        failure_message = f"FAILED to unmute @{self.username}." if mute_status \
                            else f"FAILED to mute @{self.username}."

        await self.option_handler.handle_toot_action(
            action=action,
            success_message=success_message,
            failure_message=failure_message,
            toggle_status=mute_status,
            key=self.option_handler.StatusKey.MUTING,
            on_user=True
        )

    @on(BlockUser)
    async def block_user(self) -> None:

        self.log.debug("Block user callback.")

        block_status = self.relation_dict['blocking']
        if block_status:
            await self._block_user(block_status)
        else:
            self.app.push_screen(
                BlockScreen(self.username, block_status, classes = "modal_screen"),
                self._block_user
            )

    async def _block_user(self, block_status: bool) -> None:
        
        action          = "account_unblock" if block_status else "account_block"
        success_message = f"@{self.username} unblocked." if block_status else f"@{self.username} blocked."
        failure_message = f"FAILED to unblock @{self.username}." if block_status \
                            else f"FAILED to block @{self.username}."

        await self.option_handler.handle_toot_action(
            action=action,
            success_message=success_message,
            failure_message=failure_message,
            toggle_status=block_status,
            key=self.option_handler.StatusKey.BLOCKING,
            on_user=True
        )

    ###~ ACTIONS ~###

    async def action_reply(self):
        await self.reply_to_toot()

    async def action_boost(self):
        await self.boost_toot()

    async def action_favourite(self):
        await self.favorite_toot()

    async def action_bookmark(self):
        await self.bookmark_toot()

    async def action_options(self):
        await self.show_more_options()

    def action_switch_to_tootpage(self):
        self.refresh_bindings()
        self.post_message(ExamineToot(self.toot_id))

    def check_action(
        self, action: str, parameters: tuple[object, ...]
    ) -> bool | None:  
        """Check if an action may run."""
        if action == "switch_to_tootpage":
            if self.toot_switcher.current == "toot_edit_container":
                return False
            if self.replybox_on and self.query_one("#reply_box").has_focus_within:
                return False
        return True


class TootContentContainer(Vertical):

    def __init__(self, toot_widget: TootWidget, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.toot_widget = toot_widget
        self.bs4_parser = BS4Parser()
        self.media = self.toot_widget.media_attachments
        self.image_url = self.media[0]["preview_url"] if self.media else None
        self.image_on: bool = False
        self.parsed_content = None

        # TODO Maybe should not be using preview_url for images. But it seems to work for now.
        # We can't exactly display a very high res image in a terminal anyway.

    def compose(self):
        yield SimpleButton('', id='toot_content_button', classes="toot_content", no_wrap=None, overflow=None)

        if self.app.show_images and self.image_url:
            self.image_on = True
            yield ImageViewerWidget(
                self.image_url,
                id="imgviewer_widget",
                classes="toot_image"
            )
        yield Static("", id='media_description', classes="toot_extra")

        if self.toot_widget.card:
            yield TootCardWidget(self.toot_widget, id="toot_card_widget", classes="toot_card")

    def on_mount(self):

        self.imgviewer_widget = self.query_one("#imgviewer_widget") if self.image_on else None
        self.call_after_refresh(self.load_toot_content)

    def load_toot_content(self):

        self.parsed_content = self.bs4_parser.parser(self.toot_widget.json["content"])
        self.query_one("#toot_content_button").update(self.parsed_content)
    
        if self.image_url:
            description1 = f"Image description: {self.media[0]['description']}"
            self.query_one("#media_description").update(description1)

            # There might be more than 1 attachment, but we just want the first one.
            # TODO maybe add option to load more than one image

    @on(SimpleButton.Pressed, selector='#toot_content_button')
    def switch_to_tootpage(self):
        self.post_message(ExamineToot(self.toot_widget.toot_id))


class TootOptionHandler(Widget):

    class StatusKey(str, Enum):
        """These are the keys that are checked in the returned JSON dict to see if the action was successful."""
        PINNED = "pinned"
        MUTED = "muted"             # muted is for the status / conversation. 'mute this conversation'
        REBLOGGED = "reblogged"
        FAVOURITED = "favourited"
        BOOKMARKED = "bookmarked"
        MUTING = "muting"           # muting is for the user. 'mute this user'
        BLOCKING = "blocking"

    def __init__(self, toot_widget: TootWidget) -> None:
        super().__init__()
        self.toot_widget = toot_widget

    async def handle_toot_action(
        self,
        action: str,                     # The name of the Mastodon API method to call (e.g., 'status_pin').
        success_message: str,            # Message to display on success.
        failure_message: str,            # Message to display on failure.
        toggle_status: bool,             # The ingoing status of the toggle.
        key: Optional[StatusKey] = None, # The key in the returned JSON dict to check for success.
        on_user: bool = False            # Whether the action is on the user or the status
    ) -> None:
        """Every Mastodon action passed into this method must take either a toot_id 
        or a user_id as its first argument. This is determined by the
        `on_user` parameter. If `on_user` is True, the action is on the user (toot author).
        If `on_user` is False, the action is on the status (the toot itself).
        
        'key' is the entry in the returned JSON dict that will be checked to see if the action
        was successful. If the action is successful, the value of the key should be different
        from the ingoing status. If the action is unsuccessful, the value of the key should be the same."""
        
        self.log.debug(
            f"Handling action: {action} \n"
            f"toggle_status: {toggle_status} \n"
            f"key: {key} \n"
            f"on_user: {on_user} \n"
        )
        
        if on_user:
            id_to_action = self.toot_widget.user_id    # user actions are mute, block, etc
        else:
            id_to_action = self.toot_widget.toot_id    # status actions are boost, favourite, etc

        attribute = None
        with self.app.capture_exceptions():
            attribute = getattr(self.app.mastodon, action)
        if self.app.error:
            return
        if attribute:

            status = None
            with self.app.capture_exceptions():
                status = await attribute(id_to_action)
            if self.app.error:
                return
            if status:

                self.log.debug(
                    f"{action.capitalize()} status: {status} \n"
                    f"{action.capitalize()} status isinstance of dict?:  {isinstance(status, dict)}"
                )
                if isinstance(status, dict):
                    if key and status[key] != toggle_status:  # check returned dict is different from ingoing status
                        self.post_message(SuperNotify(success_message))
                    if not key:
                        self.post_message(SuperNotify(success_message))
                else:
                    self.post_message(SuperNotify(failure_message))
                
                if action == "status_delete":
                    return
                else:
                    self.toot_widget.refresh_toot()


class TextAreaEdit(TextArea):

    class Submit(Message):
        pass

    BINDINGS = [
        Binding("ctrl+e", "submit", "Save edit", show=True),
        Binding("escape", "cancel", "Cancel edit", key_display='Esc', show=True),
        Binding("enter", "pass", "Create new line", show=False),
    ]

    def focus(self, scroll_visible: bool = True):
        """Copied from Widget.focus() but with scroll_visible set to False and removed
        the call_later thing. (it gave issues, no idea why.)"""

        try:        
            self.screen.set_focus(self, scroll_visible=False)
        except NoScreen:
            pass

        return self

    def action_submit(self):
        self.post_message(self.Submit())

    def action_cancel(self):
        self.parent.cancel_edit()

    def action_pass(self):
        pass


class TootEditContainer(Container):
    """Unlike the reply box, this is not added and removed. It's composed in the widget,
    and switched to using the ContentSwitcher."""

    def __init__(self, toot_widget: TootWidget, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.toot_widget = toot_widget
        self.content = None

    def compose(self):
        yield TextAreaEdit("", id='toot_edit_box')
        with Horizontal(classes="reply_footer"):
            yield SimpleButton("Save edit", id="toot_edit_save", classes="toot_box_button")
            yield SimpleButton("Cancel", id="toot_edit_cancel", classes="toot_box_button")

    def on_mount(self):
        self.input_box = cast(TextArea, self.query_one("#toot_edit_box"))

    def edit_toot(self) -> None:

        self.input_box.can_focus = True    # this has to be here and not in on_mount.      
        self.content = self.toot_widget.toot_content_container.parsed_content
        self.set_timer(self.app.text_insert_time, self.set_content)   # to solve text glitching

    def set_content(self) -> None:
        self.input_box.text = self.content
        self.input_box.focus()
        self.input_box.action_cursor_line_end()
        self.app.post_message(ScrollToWidget(self.input_box))


    @on(TextAreaEdit.Submit)
    @on(SimpleButton.Pressed, selector="#toot_edit_save")
    async def save_edit(self) -> None:
        # I did this because the arguments here are too different
        # from the normal option handler to justify using it.

        with self.app.capture_exceptions():
            await self.app.mastodon.status_update(
                id=self.toot_widget.toot_id,
                status=self.input_box.text,
                spoiler_text=None,                      #! TODO implement these
                sensitive=None,                         # Roadmap: More tooting options
                media_ids=None,
                poll=None
            )
        
        self.toot_widget.switch_content("toot_content_container")
        if not self.app.error:
            self.toot_widget.refresh_toot()

    @on(SimpleButton.Pressed, selector="#toot_edit_cancel")
    def cancel_edit(self) -> None:
        self.toot_widget.switch_content("toot_content_container")
        self.toot_widget.edit_mode = False
        self.toot_widget.focus()


class TootCardWidget(Vertical):

    def __init__(self, toot_widget: dict, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.toot_widget = toot_widget
        self.card_json = toot_widget.card

        self.url        = f"{self.card_json['url']} \n"
        self.card_title = f"{self.card_json['title']} \n"
        self.image_url  =  self.card_json["image"]
        self.image_description = self.card_json["image_description"]        

        img_desc:str = shorten(self.card_json["description"], width=100)

        if self.app.link_behavior == 0:
            self.tooltip = f"{img_desc} \n\n Click to open in browser"
        elif self.app.link_behavior == 1:
            self.tooltip = f"{img_desc} \n\n Click to copy link to clipboard"
        elif self.app.link_behavior == 2:
            self.tooltip = f"{img_desc} \n\n Click to open link helper popup"

    def compose(self):

        # If there's both media attachments and a toot card, we only want to show the
        # media attachments. Because otherwise its just too much image.

        if self.app.show_images and self.image_url and not self.toot_widget.media_attachments:
            yield ImageViewerWidget(self.image_url, in_card=True, classes="toot_image")
        yield Static("", id="toot_card_title", classes="toot_card_data")
        yield Static("", id="toot_card_url", classes="toot_card_data")
        yield Static("", id="toot_card_image_desc", classes="toot_card_data")

    def on_mount(self):
        
        # This uses the Rich Text class to make the URL crop if it's too long
        rich_url = Text(self.url, no_wrap=True, overflow="crop")
        self.query_one("#toot_card_url").update(rich_url)

        self.query_one("#toot_card_title").update(self.card_title)
        if self.image_description:
            self.query_one("#toot_card_image_desc").update(f"Image description: {self.image_description}")

    def on_click(self) -> None:

        self.app.handle_link(self.card_json["url"])

    # This is a trick to make it light up the entire card and all children on hover.
    def on_enter(self):
        self.styles.background = self.app.theme_variables["panel-lighten-1"]

    def on_leave(self):
        self.styles.background = self.app.theme_variables["surface-lighten-1"]








# Standard library imports
from __future__ import annotations
from typing import TYPE_CHECKING, cast, List, Tuple

if TYPE_CHECKING:
    from textual.app import ComposeResult 

# Third party imports
from mastodon import Mastodon

# Textual imports
from textual import on, work
from textual.worker import Worker, WorkerState
from textual.binding import Binding
from textual.message import Message
from textual.containers import Container, Horizontal
from textual.widget import Widget
from textual.widgets import Static
# from textual.reactive import reactive
from rich.text import Text
from rich.emoji import Emoji

# TextualDon imports
from textualdon.simplebutton import SimpleButton
from textualdon.messages import (
    SuperNotify,
    LoginStatus,
)
from textualdon.sql import SQLite    # only for casting
from textualdon.screens import ConfirmationScreen


class UserEntry(Horizontal):

    BINDINGS = [
        Binding(key="delete", action="delete_user", description="Delete highlighted user", key_display="Delete"),
    ]
    class UserSelected(Message):
        def __init__(self, user: UserEntry) -> None:
            super().__init__()
            self.user_entry = user

    class UserDeleted(Message):
        def __init__(self, user: UserEntry) -> None:
            super().__init__()
            self.user_entry = user

    def __init__(
            self,
            user_id: int,
            instance_url: str,
            username: str,
            display_name: str,
            access_token: str,
            *args,
            **kwargs
        ) -> None:
        self.user_id = user_id
        self.instance_url = instance_url
        self.username = username
        self.display_name = display_name
        self.access_token = access_token
        super().__init__(*args, **kwargs)

    def compose(self) -> ComposeResult:

        yield SimpleButton(
            f"{self.display_name} (@{self.username}) \n{self.instance_url}",
            # no_wrap=False,
            id="user_select_button"
        )
        yield SimpleButton(f" Delete {Emoji('wastebasket')} ", id="user_delete_button")

    def on_mount(self):
        self.user_select_button = self.query_one("#user_select_button")
        self.user_delete_button = self.query_one("#user_delete_button")
        self.user_delete_button.can_focus = False

        self.log.debug(
            "UserEntry mounted with: \n"
            f"User ID: {self.user_id} \n"
            f"Instance URL: {self.instance_url} \n"
            f"Username: {self.username} \n"
            f"Display Name: {self.display_name} \n"
            f"Access Token: {self.access_token}"
        )

    @on(SimpleButton.Pressed, selector="#user_select_button")
    def user_selected(self) -> None:
        self.log.debug("Messaging function: user_selected in UserEntry class (SimpleButton.Pressed)")
        self.post_message(self.UserSelected(self))

    @on(SimpleButton.Pressed, selector="#user_delete_button")
    def delete_user(self) -> None:
        self.log("Messaging function: delete_user in UserEntry class (SimpleButton.Pressed)")
        self.post_message(self.UserDeleted(self))

    def action_delete_user(self) -> None:
        self.delete_user()


class SavedUsersManager(Widget):
    """Composed onto the Oauth widget to manage saved users. \n
    DOM id = 'saved_users_manager' """

    users_list = []

    class TriggerLogin(Message):
        """Message to trigger the login_stage5 function."""
        def __init__(self, instance_url: str) -> None:
            super().__init__()
            self.instance_url = instance_url


    def compose(self) -> ComposeResult:

        self.db = cast(SQLite, self.app.sqlite)

        with Container(id="users_container"):
            yield Static("No saved logins", classes="users_placeholder")

    def on_mount(self):

        self.users_container = cast(Container, self.query_one("#users_container"))
        self.users_container.border_title = "Saved Logins" 
            
    async def start_process(self):
        """This is activated by the on_mount method in Oauth widget, as well as the 
        user_deleted function in this class to refresh the list."""

        with self.app.capture_exceptions():
            users_list = await self.get_saved_logins()
        if self.app.error:
            return
        else:
            self.log.debug("Saved users list successfully retrieved.")
            if users_list:
                self.users_list = users_list
                await self.mount_saved(users_list)

    async def get_saved_logins(self) -> List[UserEntry]:

        self.log("Starting get_saved_logins in OAuthWidget")

        query = "SELECT * FROM users"
        users = self.db.fetchall(query)     #~ This is the point we query the database to get all the user data.

        users_list = []

        if users:
            self.log(Text(f"Users found in database: {len(users_list)}", style="green"))
            for user in users:
                user_id = user[0]
                instance_url = user[1]
                username = user[2]
                display_name = user[3]
                access_token = user[4]

                user_entry = UserEntry(user_id, instance_url, username, display_name, access_token)
                users_list.append(user_entry)
        return users_list

    async def mount_saved(self, users_list: List[UserEntry]) -> None:

        await self.users_container.remove_children()
        self.users_container.mount_all(users_list)
        self.users_container.refresh(layout=True)

    def check_auto_login(self) -> None:
        """This is the trigger function used by App. This is run at the end of App's on_mount function."""
        if self.app.safe_mode:
            return

        row1 = self.db.fetchone("SELECT value FROM settings WHERE name = ?", ("last_logged_in",))
        last_logged_in: int = int(row1[0])

        row2 = self.db.fetchone("SELECT value FROM settings WHERE name = ?", ("auto_login",))
        autologin_value: bool = (row2[0] == "True")

        if autologin_value and self.users_list:
            for user in self.users_list:
                if user.user_id == last_logged_in:
                    self.post_message(user.UserSelected(user))

    @work(thread=True, group="create_mastodon_instance", exit_on_error=False)
    def create_mastodon_instance(
            self, 
            access_token:str, 
            instance_url:str, 
            app_data:Tuple[str, str]
        ) -> Mastodon:
        """Used by user_selected and user_deleted functions to create a Mastodon object."""

        try:        
            mastodon = Mastodon(                #~ Creates Mastodon instance here.
                client_id = app_data[2],
                client_secret = app_data[3],
                access_token = access_token,
                api_base_url = instance_url,
                version_check_mode='none',
                request_timeout=3
            )          
        except Exception as e:
            e.add_note("SavedUsersManager.create_mastodon_instance failed to create Mastodon object.")
            raise e
        else:
            return mastodon

    @on(UserEntry.UserSelected)
    async def user_selected(self, event: UserEntry.UserSelected) -> None:

        if self.app.safe_mode:
            self.notify("Please disable safe mode first.", timeout=7)
            return        

        access_token: str = event.user_entry.access_token
        instance_url: str = event.user_entry.instance_url

        app_data = self.db.fetchone("SELECT * FROM app_data WHERE instance_url = ?", (instance_url,))
        
        self.log.debug(
            "Starting user_selected in OAuthWidget.\n"
            f"app_data: \n{app_data} \n"
            f"access_token: {access_token} \n"
            f"instance_url: {instance_url}"
        )

        with self.app.capture_exceptions():
            worker = self.create_mastodon_instance(access_token, instance_url, app_data)
            await worker.wait()
            mastodon = worker.result
        if not self.app.error:
            self.app.attach_mastodon(mastodon)                      # we're already logged in, 
            self.post_message(self.TriggerLogin(instance_url))      # no need to pass in access token      

    @on(UserEntry.UserDeleted)
    def user_deleted_confirm(self, event: UserEntry.UserDeleted) -> None:
        if self.app.safe_mode:
            self.notify("Please disable safe mode first.", timeout=7)
            return

        self.app.push_screen(ConfirmationScreen(forward=event, classes="modal_screen"), self.user_deleted)

    async def user_deleted(self, event: UserEntry.UserDeleted) -> None:
        """Triggered by the callback from the ConfirmationScreen."""

        access_token = event.user_entry.access_token
        instance_url = event.user_entry.instance_url
        username = event.user_entry.username
        user_id = event.user_entry.user_id

        self.log.debug(
            "Starting user_deleted function in OAuthWidget. Parameters: \n"
            f"access_token: {access_token} \n"
            f"instance_url: {instance_url} \n"
            f"username: {username} \n"
            f"user_id: {user_id}"
        )

        # If deleting the logged in user, log out first. It's easier to just make it
        # re-validate, because it simplifies the logic.

        # NOTE: self.app.logged_in_user_id is set in login_stage5 in Oauth widget.

        if self.app.logged_in_user_id:
            if user_id == self.app.logged_in_user_id:   
                self.app.mastodon = None               
                self.app.logged_in_user_id = None
                self.post_message(LoginStatus(
                    statusbar="Status: Offline",
                    loginpage_message="Logged out."
                    ))
                self.post_message(SuperNotify("Logged out."))

        app_data = self.db.fetchone("SELECT * FROM app_data WHERE instance_url = ?", (instance_url,))

        worker = self.create_mastodon_instance(access_token, instance_url, app_data)
        await worker.wait()
        mastodon = worker.result
        if not self.app.error:   
            with self.app.capture_exceptions(): 
                mastodon.revoke_access_token()
            if self.app.error:
                return

        self.log(f"Mastodon access token revoked for {username} on {instance_url}")

        self.db.delete_one('users', 'id', user_id)
        await self.start_process()
        self.post_message(SuperNotify(f"Deleted user {username} on {instance_url}"))

    ###~ Worker Debug stuff ~###

    @on(Worker.StateChanged)
    def worker_state_changed(self, event: Worker.StateChanged) -> None:

        if event.state == WorkerState.SUCCESS:
            self.log(Text(f"Worker {event.worker.name} completed successfully", style="green"))
        elif event.state == WorkerState.ERROR:
            self.log.error(Text(f"Worker {event.worker.name} encountered an error", style="red"))
        elif event.state == WorkerState.CANCELLED:
            self.log(Text(f"Worker {event.worker.name} was cancelled", style="yellow"))




# Standard Library imports
from __future__ import annotations
from typing import TYPE_CHECKING, cast, List, Tuple
import urllib.request
from datetime import datetime


if TYPE_CHECKING:
    from textual.app import ComposeResult 

# Third party imports
from mastodon import Mastodon
from rich.text import Text
import PIL.Image

# Textual imports
from textual import on, work
# from textual.reactive import reactive
# from textual.dom import NoScreen
from textual.worker import Worker, WorkerState
from textual.binding import Binding
from textual.message import Message
from textual.containers import Horizontal, Vertical, Container
from textual.widget import Widget
from textual.widgets import (
    Sparkline,
    Pretty,
    Static,
    # TextArea,
    Input,
)

# TextualDon imports
from textualdon.screens import ImageScreen, MessageScreen, NotImplementedScreen
from textualdon.simplebutton import SimpleButton
# from textualdon.messages import ScrollToWidget
from textualdon.imageviewer import ImageViewer    # takes PIL.Image.Image as only argument


class InputCustom(Input):

    BINDINGS = [
        Binding("enter", "submit", "Submit", key_display='Enter', show=True),
    ]


class WelcomeWidget(Container):

    welcome_text = """TextualDon is designed for both keyboard and mouse users. \
The bar at the bottom of the screen will show you the available commands for whatever you have focused. \n
You can see your status in the top right corner of the screen. \
Hover your mouse over it to see which instance you are connected to (or check this page). \n
You can also use the command palette (bottom right, or press 'ctrl + p') to read Textual's command list \
and to change themes. (I tried to make sure TextualDon looks good in most themes, and \
it will remember your theme setting as well.) \n"""
    welcome_text2 = """Note that default link behavior is to open in your browser. \
You can change this in the settings below. This will be useful for people who cannot \
open a browser window automatically."""
    alpha_text = """Hey there! This program is still in alpha stages. If you're reading this, \
you're probably one of the first people to try it. Many features are still missing, and \
there's a chance you'll run into bugs. If you do, please let me know! The reporting \
screen should save you some time in doing so, assuming it works properly. \n"""

    def compose(self):

        with Container(id="text", classes="page_box content"):
            yield SimpleButton("Testers/Discord group press here", id="testers_button", classes="page_button short")
            yield Static(self.welcome_text, classes="page_box content")
        with Horizontal(classes="page_box bar"):
            yield Static(self.welcome_text2, id="welcome2", classes="page_box content")
            yield SimpleButton("Hide", id="hide_button", classes="page_button bordered")

    def on_mount(self):

        self.query_one("#hide_button").can_focus = True

        row = self.app.sqlite.fetchone(
            "SELECT value FROM settings WHERE name = ?", ("show_welcome_message",)
        )
        show_welcome = (row[0] == "True")
        if not show_welcome:
            self.hide_widget(update_db=False)

    @on(SimpleButton.Pressed, selector="#hide_button")
    def show_hide_trigger(self) -> None:

        if self.query_one("#text").display is True:
            self.hide_widget()
        else:
            self.show_widget()

    @on(SimpleButton.Pressed, selector="#testers_button")
    def show_alpha_message(self) -> None:

        self.app.push_screen(MessageScreen(self.alpha_text, classes="modal_screen"))

    def hide_widget(self, update_db: bool = True) -> None:

        self.query_one("#text").display = False
        self.query_one("#hide_button").update("Show")
        self.query_one("#welcome2").update("\nPress 'Show' to see the introduction again.")
        self.set_classes("page_box message hidden")

        if update_db:
            self.app.sqlite.update_column("settings", "value", "False", "name", "show_welcome_message")

    def show_widget(self, update_db: bool = True) -> None:

        self.query_one("#text").display = True
        self.query_one("#hide_button").update("Hide")
        self.query_one("#welcome2").update(self.welcome_text2)
        self.set_classes("page_box message")

        if update_db:
            self.app.sqlite.update_column("settings", "value", "True", "name", "show_welcome_message")


class TimelineSelector(Widget):

    BINDINGS = [
        Binding("left", "focus_previous", "Focus left", show=True),
        Binding("right", "focus_next", "Focus right", show=True),
    ]
    # focused_index = reactive(0)     # start on leftmost timeline

    class ChangeTimeline(Message):
        """This message is sent when a timeline is selected in the TimelineSelector widget.
        This is handled in the page the TimelineSelector is mounted in."""
        def __init__(self, timeline: str) -> None:
            super().__init__()
            self.timeline = timeline

    def __init__(self, options: List[Tuple[str, str]], **kwargs):
        super().__init__(**kwargs)
        self.options = options
        self.current = 0        # index of the selected timeline

        # NOTE: Every entry in options is a tuple of the display name and id of the timeline.
        self.buttons_list = [
            SimpleButton(option[0], id=option[1], index=index, classes="timeline_button") 
            for index, option in enumerate(self.options)
        ]
        self.amt_of_buttons = len(self.buttons_list)

    def compose(self) -> ComposeResult:

        #~ NOTE: Timeline CSS is in pages.tcss

        with Horizontal(id="timeline_container", classes="page_box timeline"):
            for button in self.buttons_list:
                yield button

    def on_mount(self):
        timeline_container = self.query_one("#timeline_container")
        self.buttons = list(timeline_container.query_children().results())

        selected = self.query_one(f"#{self.options[self.current][1]}")
        selected.set_classes("timeline_button selected")

    @on(SimpleButton.Pressed)
    def switch_timeline(self, event: SimpleButton.Pressed) -> None:

        if event.button.index == self.current:
            self.log.debug("Already on this timeline.")
            return
        
        self.log.debug(f"Switching to {event.button.id}")
        for option in self.options:                         # First reset all buttons
            button = self.query_one(f"#{option[1]}")
            button.set_classes("timeline_button")

        selected = self.query_one(f"#{event.button.id}")
        selected.set_classes("timeline_button selected")
        self.current = event.button.index
        self.post_message(self.ChangeTimeline(event.button.id))  # Send message to parent page

    def action_focus_previous(self) -> None:
        if self.buttons[0].has_focus:
            self.buttons[-1].focus()
        self.screen.focus_previous()

    def action_focus_next(self) -> None:
        if self.buttons[-1].has_focus:
            self.buttons[0].focus()
        self.screen.focus_next()


class MiscMastoWidget(Widget):


    def on_focus(self):
        self.log.debug(f"{self.name} focused. ")
        self.styles.border = ('dashed', self.app.theme_variables["primary"])

    def on_blur(self):
        self.styles.border = ('blank', 'transparent')


    def get_days_of_week(self):
        # Get the current day of the week
        current_day = datetime.now().strftime("%a")
        day_list = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"]
        day_codes = ["M", "T", "W", "T", "F", "S", "S"]
        index = day_list.index(current_day)
        if index == 6:
            return " ".join(day_codes)
        else:
            my_days = day_codes[index+1:] + day_codes[:index+1]
            return " ".join(my_days)


class HashtagWidget(MiscMastoWidget):

    BINDINGS = [
        Binding("enter", "switch_to_hashtagpage", "Expand hashtag", show=True),
    ]

    def __init__(self, json: dict, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.json = json

    def compose(self):
        with Container(classes="content_container"):
            yield SimpleButton("hashtag", id="hashtag_name", classes="titlebar")
            with Horizontal(classes="trend_footer"):
                yield Static("history", id="hashtag_history", classes="trend_footer nums")
                with Vertical(classes="trend_footer sparkline"):
                    yield Sparkline([0], id="hashtag_sparkline", classes="sparkline")
                    yield Static(self.get_days_of_week(), classes="trend_label")

    def on_mount(self):

        self.can_focus = True
        self.query_one("#hashtag_name").can_focus = False
        history = self.json["history"]
        counts_list, past_2_days, past_week = self.app.get_history_data(history)

        self.query_one("#hashtag_name").update(f"#{self.json['name']}")
        self.query_one("#hashtag_name").tooltip = self.json["url"]
        self.query_one("#hashtag_history").update(
                    f"{past_2_days} people in the past 2 days. \n"
                    f"{past_week} people in the past week.")
        self.query_one("#hashtag_sparkline").data = counts_list

        self.loading = False

    def action_switch_to_hashtagpage(self):
        self.app.push_screen(NotImplementedScreen("More pages"))


class NewsWidget(MiscMastoWidget):

    BINDINGS = [
        Binding("enter", "switch_to_newspage", "Expand news story", show=True),
    ]

    def __init__(self, json: dict, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.json = json
        self.html_on = False     # flag
        self.json_on = False     # flag
        self.loading = True

    # TODO These should be using the card widget for the news.

    def compose(self):
        with Container(classes="content_container"):
            yield Static("title", id="news_title", classes="titlebar")
            yield Static("url", id="news_url")
            yield Static("description", id="news_description")
            yield Static("author", id="news_author")
            yield Static("date", id="news_date")
            yield Static("provider", id="news_provider")

            with Horizontal(classes="trend_footer news"):
                yield Static("history", id="news_history", classes="trend_footer nums")
                with Vertical(classes="trend_footer sparkline"):
                    yield Sparkline([0], id="news_sparkline", classes="sparkline")
                    yield Static(self.get_days_of_week(), classes="trend_label")

    def on_mount(self):

        self.can_focus = True
        counts_list, past_2_days, past_week = self.app.get_history_data(self.json["history"])
        date_object: type = self.json["published_at"]  # Mastodon.py returns datetime objects
        self.log.debug(date_object)
        self.log.debug(date_object.__class__)
        self.log.debug(date_object.__class__.mro())
        
        self.query_one("#news_title").update(self.json["title"])
        self.query_one("#news_url").update(self.json["url"])
        self.query_one("#news_description").update(self.json["description"])
        self.query_one("#news_author").update(self.json["author_name"])
        # self.query_one("#news_date").update(date)
        self.query_one("#news_provider").update(self.json["provider_name"])
        self.query_one("#news_history").update(
                    f"{past_2_days} people in the past 2 days. \n"
                    f"{past_week} people in the past week.")
        self.query_one("#news_sparkline").data = counts_list

        self.loading = False

    def action_switch_to_newspage(self):
        self.app.push_screen(NotImplementedScreen("More pages"))


# TODO Make PeopleWidget


class ImageViewerWidget(Container):
    """This is a simple widget that displays an image in a container.
    It's used across the program to display images. It's controlled by the ImageViewer class.
    
    Connected to screens.ImageScreen for fullscreen viewing."""

    def __init__(self, image_url: str, in_card: bool = False, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.url = image_url
        self.in_card = in_card  # If True, the image is displayed in a card.
        self.can_focus = False

    async def on_mount(self):
        
        with self.app.capture_exceptions():
            img_worker = self.load_image_from_url()
            self.img = await img_worker.wait()
        if self.app.error:
            return

        self.imgview = ImageViewer(self.img, nested=True, id="imgview")
        self.mount(self.imgview)
        if not self.in_card:
            self.tooltip = 'Click to view full size'

    @work(thread=True, exit_on_error=False, group="image")
    async def load_image_from_url(self) -> PIL.Image.Image:

        headers = {"User-Agent": "Mozilla/5.0"}

        try:
            req = urllib.request.Request(self.url, headers=headers)
        except Exception as e:
            raise e
        
        with urllib.request.urlopen(req) as response:
            return PIL.Image.open(response)
        
    async def on_click(self):
        if not self.in_card:
            await self.fullscreen()

    async def fullscreen(self):
        await self.app.push_screen(ImageScreen(self.img, classes="fullscreen"))

    @on(Worker.StateChanged)
    def worker_state_changed(self, event: Worker.StateChanged) -> None:
        
        if event.state == WorkerState.SUCCESS:
            self.log(Text(f"Worker {event.worker.name} completed successfully", style="green"))
        elif event.state == WorkerState.ERROR:
            self.log.error(Text(f"Worker {event.worker.name} encountered an error", style="red"))
        elif event.state == WorkerState.CANCELLED:
            self.log(Text(f"Worker {event.worker.name} was cancelled", style="yellow"))


class ProfileWidget(MiscMastoWidget):
    """Displays a user's profile. Used by the UserProfilePage class."""

    def __init__(self, account_dict, relation_dict, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.account_dict = account_dict
        self.relation_dict = relation_dict

    def compose(self):
        yield Pretty(self.account_dict, classes="page_box")
        yield Pretty(self.relation_dict, classes="page_box")

    def on_mount(self):
        self.mastodon = cast(Mastodon, self.app.mastodon)



# Standard Library Imports
from __future__ import annotations
from typing import cast, Any
import time
from pathlib import Path
from importlib import resources

# Third party imports
import clipman
import pyperclip
import PIL.Image

# Textual imports
from textual import on
from textual.binding import Binding
from textual.screen import Screen, ModalScreen
from textual.containers import Container, Horizontal, VerticalScroll
from textual.widgets import Button, Label, Checkbox, TextArea, Markdown

# TextualDon imports
from textualdon.simplebutton import SimpleButton
from textualdon.messages import CallbackCancel, OpenRoadmap
from textualdon.sql import SQLite       # this is only for casting purposes
from textualdon.imageviewer import ImageViewer

class TextualdonModalScreen(ModalScreen):

    BINDINGS = [
        Binding("escape", "pop_screen", key_display='Esc', description="Close the pop-up screen.", show=True),
        Binding("up,left", "focus_previous", description="Focus the previous button."),
        Binding("down,right", "focus_next", description="Focus the next button."),
    ]

    controls = "Arrow keys (or tab): navigate | Enter: select | Esc: close"

    def action_pop_screen(self):

        self.log.info(f"screen stack: {self.app.screen_stack}")
        self.app.pop_screen()
        # screen_stack = self.app.screen_stack

    def on_mount(self):
        self.mount(
            Container(
                Label(self.controls, classes='screen_label'),
                classes='screen_container wide help',
                id="help_container"
            )
        )        

    def action_focus_previous(self):
        self.focus_previous()
    
    def action_focus_next(self):
        self.focus_next()


class ImageScreen(Screen):
    """Called by `ImageViewerWidget.fullscreen` (widgets.py)
    Callback: None."""

    BINDINGS = [
        Binding("escape", "dismiss", key_display='Esc', description="Close the image screen."),
        Binding("i", "zoom_in", description="Zoom the image in."),
        Binding("o", "zoom_out", description="Zoom the image out."),
        Binding("up", "pan_up", description="Pan the image up."),
        Binding("down", "pan_down", description="Pan the image down."),
        Binding("left", "pan_left", description="Pan the image left."),
        Binding("right", "pan_right", description="Pan the image right."),
    ]

    def __init__(self, image: PIL.Image.Image, **kwargs):
        self.image = image
        super().__init__(**kwargs)

    def compose(self):
        with Container(id='imgview_container', classes='fullscreen'):
            yield ImageViewer(self.image) 
        with Horizontal(classes='screen_buttonbar'):
            yield Button('Close', id='img_close_button')
            yield Label('i/o and arrows (or mousewheel/drag): Zoom in/out and pan | Esc to close', id='imgview_label')

    def on_mount(self):
        self.img_container = self.query_one('#imgview_container')
        self.img_viewer = self.query_one(ImageViewer)
        self.img_container.can_focus = True         # TODO test if this is necessary

    def on_button_pressed(self, button):
        self.dismiss()

    def action_zoom_in(self):
        self.log.info('Zooming in')
        self.img_viewer.image.zoom(-1)
        self.img_viewer.refresh()

    def action_zoom_out(self):
        self.img_viewer.image.zoom(1)
        self.img_viewer.refresh()

    def action_pan_up(self):
        self.img_viewer.image.move(0, 2)    # NOTE: the integer values here are actually
        self.img_viewer.refresh()           # reveresed from how its done in the ImageViewer class.

    def action_pan_down(self):                  # That's because with mouse drag, you always drag up
        self.img_viewer.image.move(0, -2)       # to pull the image down, and vice versa.
        self.img_viewer.refresh()               # But with keyboard, its not normally reversed.

    def action_pan_left(self):                  # So we want up key to move the image up, and
        self.img_viewer.image.move(2, 0)        # Down key to move the image down.
        self.img_viewer.refresh()

    def action_pan_right(self):                 # Also using 2 instead of 1 to make it more sensitive.
        self.img_viewer.image.move(-2, 0)
        self.img_viewer.refresh()


class ConfirmationScreen(TextualdonModalScreen):
    """ Generic screen used in two places. \n
    Called by:
    - `TootWidget.delete_toot` | Callback: `TootWidget._delete_toot` (toot.py) 
    - `SavedUsersManager.user_deleted_confirm` | Callback: `SavedUsersManager.user_deleted` (savedusers.py) """

    def __init__(
        self, 
        forward: Any = None, 
        **kwargs
    ):
        self.forward = forward
        super().__init__(**kwargs)

    def compose(self):
        with Container(classes='screen_container'):
            yield Label('Are you sure? \n', classes='screen_label')
            with Horizontal(classes='screen_buttonbar'):
                yield Button('Yes', id='confirm_yes_button')
                yield Button('Cancel', id='confirm_no_button')

    @on(Button.Pressed, selector='#confirm_yes_button')
    async def confirm_yes(self):

        self.log.info(f'Forward value: {self.forward}')
        if self.forward:
            self.dismiss(self.forward)
        else:
            self.dismiss()         

    @on(Button.Pressed, selector='#confirm_no_button')
    def confirm_no(self):
        self.action_pop_screen()


class NotImplementedScreen(TextualdonModalScreen):
    """ Generic screen used in three places. | Callbacks: None \n
    Called by:
    - `TootBox.search_mode` (tootbox.py)
    - `TootOptionsOtherUser.report_user` (tootscreens.py) 
    - `TootOptionsOtherUser.filter_toot` (tootscreens.py)"""

    controls = "Arrow keys (or tab): navigate | Enter: select | Esc or click anywhere: close"

    def __init__(self, roadmap_name: str, **kwargs):
        super().__init__(**kwargs)
        self.roadmap_name = roadmap_name

    def compose(self):
        with Container(classes='screen_container'):
            yield Label((
                'This feature is not yet implemented. \n\n'
                f'Roadmap entry: {self.roadmap_name} \n'
                ), classes='screen_label'
            )
            with Horizontal(classes='screen_buttonbar'):
                yield Button('Close', id='close_button')
                yield Button('View Roadmap', id='roadmap_button')               

    def on_click(self):
        self.dismiss()

    @on(Button.Pressed, selector='#close_button')
    def report_close(self):
        self.dismiss()  

    @on(Button.Pressed, selector='#roadmap_button')
    def roadmap_button(self):
        self.app.post_message(OpenRoadmap())
        self.dismiss()


class WSLWarning(TextualdonModalScreen):
    """Intro / First time user screen.
    Called by `app.on_mount` | Callback: `app.intro_screens_callback` """

    BINDINGS = [Binding("escape", "pass", show=False)]   # remove the escape key feature here
    def action_pass(self):
        pass
    controls = "Arrow keys (or tab): navigate | Enter: select"

    wsl_warning = """TextualDon has detected that it's running inside of WSL. \n
You might notice little flashes of an error whenever you open a link in your browser. \n
Unfortunately there's nothing I can do about that. You can just ignore it. \n
If you know a solution, feel free to let me know on the github page. \n"""
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.db = cast(SQLite, self.app.sqlite)

    def compose(self):

        with Container(classes='screen_container wide'):
            yield Label(self.wsl_warning, classes='screen_label')
            with Horizontal(classes='screen_buttonbar'):
                yield Checkbox("Don't show again", id='warning_checkbox')
                yield Button('Close', id='close_button')               

    def on_mount(self):
        self.checkbox = self.query_one('#warning_checkbox')
        self.focus_next()   

    @on(Button.Pressed, selector='#close_button')
    def report_close(self):
        self.dismiss()    # callback: intro_screens_callback on App

    @on(Checkbox.Changed, selector='#warning_checkbox')
    def toggle_checkbox(self, event: Checkbox.Changed):

        self.db.update_column('settings', 'value', str(event.value), 'name', 'warning_checkbox_wsl')


class FirstWarning(TextualdonModalScreen):
    """Intro / First time user screen.
    Called by `app.on_mount` | Callback: `app.intro_screens_callback` """

    BINDINGS = [Binding("escape", "pass", show=False)]   # remove the escape key feature here
    def action_pass(self):
        pass
    controls = "Arrow keys (or tab): navigate | Enter: select"

    first_warning = """First time users: \n
The Mastodon authentication process will require you to open a browser window \
at least once to login. \n
Unfortunately this does not work over SSH yet. It is on the roadmap. \n
There's 3 options provided to get links into your browser. If one of them \
isn't working, try a different option. \n"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.db = cast(SQLite, self.app.sqlite)

    def compose(self):

        with Container(classes='screen_container wide'):
            yield Label(self.first_warning, classes='screen_label')
            with Horizontal(classes='screen_buttonbar'):
                yield Checkbox("Don't show again", id='warning_checkbox')
                yield Button('Close', id='close_button')    

    def on_mount(self):
        self.focus_next()

    @on(Button.Pressed, selector='#close_button')
    def report_close(self):
        self.dismiss()   # callback: intro_screens_callback on App

    @on(Checkbox.Changed, selector='#warning_checkbox')
    def toggle_checkbox(self, event: Checkbox.Changed):

        self.db.update_column('settings', 'value', str(event.value), 'name', 'warning_checkbox_first')


class CallbackScreen(TextualdonModalScreen):
    """ Called by `OauthWidget.login_stage3` for the Oauth callback.
    Callback function: None. """
    
    # override the escape key feature here
    BINDINGS = [Binding("escape", "action_cancel_callback", description="Cancel Login", show=False)]   

    controls = "Arrow keys (or tab): navigate | Enter: select | Esc: cancel"

    def __init__(self, link: str, **kwargs):
        super().__init__(**kwargs)
        self.link = link
        self.link_button: SimpleButton | None = None
        self.callback_wait_time = self.app.config.getint('MAIN', 'callback_wait_time')

        if self.app.link_behavior == 0:     
            self.mode_msg = 'Current mode: Open browser window'
            self.label_msg = (
                '[green]Opening browser window.[/green]\n\n'
                'If needed, use an alternative option below.'
            )
        elif self.app.link_behavior == 1:  
            self.mode_msg = 'Current mode: Copy to clipboard'
            self.label_msg = (
                '[green]Link copied to clipboard.[/green]\n'
                'Paste the link in your browser to continue.\n\n'
                'If needed, use an alternative option below.'
            )
        elif self.app.link_behavior == 2:
            self.mode_msg = 'Current mode: Manual copy'
            self.label_msg = (
                'If needed, use an alternative option below.\n'
            )

    def compose(self):
        with Container(classes='screen_container wide'):
            yield Label(self.mode_msg, classes='screen_label')
            yield Label(self.label_msg, classes='screen_label')
            yield TextArea(id="link_box", read_only=True, classes="link_box")
            yield SimpleButton("Open in browser", id='browser_button', classes='screen_button')
            yield SimpleButton("Copy to clipboard", id='clipboard_button', classes='screen_button')
            yield Label(id='callback_label', classes='screen_label')                
            with Horizontal(classes='screen_buttonbar'):
                yield Button('Cancel', id='cancel_button')                

    def on_mount(self):

        self.callback_label   = self.query_one('#callback_label')
        self.browser_button   = self.query_one('#browser_button')
        self.clipboard_button = self.query_one('#clipboard_button')

        def insert_text():
            self.query_one('#link_box').text = self.link
        self.set_timer(self.app.text_insert_time, insert_text)

        if self.app.link_behavior != 2:
            self.app.handle_link(self.link)

        self.start_time = time.time()
        self.set_interval(1, self.update_countdown, repeat=self.callback_wait_time)

    @on(Button.Pressed, selector='#cancel_button')
    def action_cancel_callback(self):
        self.post_message(CallbackCancel())
        self.dismiss()

    @on(SimpleButton.Pressed, selector='#browser_button')
    def link_browser(self):

        def revert_button():
            self.browser_button.update("Open in browser")

        self.app.open_browser(self.link)
        self.browser_button.update('Link opened in browser.')
        self.set_timer(2, revert_button)

    @on(SimpleButton.Pressed, selector='#clipboard_button')
    def link_clipboard(self):

        def revert_button():
            self.clipboard_button.update("Copy to clipboard")

        self.app.copy_to_clipboard(self.link)
        self.clipboard_button.update('Link copied to clipboard.')
        self.set_timer(2, revert_button)

    def update_countdown(self):
        seconds = int(self.callback_wait_time - (time.time() - self.start_time))
        self.callback_label.update(f"Timeout in {seconds} seconds.")


class LinkScreen(TextualdonModalScreen):
    """Called by `App.handle_link` when a link is clicked in Manual Copy mode.
    Callback: None."""

    def __init__(self, link: str, **kwargs):
        super().__init__(**kwargs)
        self.link = link

    def compose(self):
        with Container(classes='screen_container'):
            yield Label('You can copy the link below with ctrl-C (or your normal command)', classes='screen_label')
            yield TextArea(id="link_box", classes="link_box", read_only=True)
            with Horizontal(classes='screen_buttonbar'):
                yield Button('Close', id='close_button')

    def on_mount(self):

        def insert_text():
            self.query_one('#link_box').text = self.link
        self.set_timer(self.app.text_insert_time, insert_text)

    @on(Button.Pressed, selector='#close_button')
    def report_close(self):
        self.dismiss()


class CopyPasteTester(TextualdonModalScreen):
    """Called by `Settings.open_tester_screen` (settings.py)
    Callback: None."""

    instructions = """This will help determine which copy/paste engine works best
on your system. \n
Use a program like Notepad to write some text, and copy \
that text to your clipboard. Then write the exact same text in the box below, \
(or try to paste it), and press the test button. \n
The app will attempt to access your copy-paste buffer and see if it matches \
what you entered.

If above you see 'Internal test failed' on either one, There's a good chance it won't \
work, and you should use one of the other options. \n
Note the app will attempt to use whatever engine you set, regardless of the test results.\n"""

    def compose(self):

        if self.app.clipman_works:
            clipman_status = "Clipman: [green]Internal test passed[/green]\n"
        else:
            clipman_status = "Clipman: [red]Internal test failed[/red]\n"
        if self.app.pyperclip_works:
            pyperclip_status = "Pyperclip: [green]Internal test passed[/green]\n"
        else:
            pyperclip_status = "Pyperclip: [red]Internal test failed[/red]\n"

        if self.app.copypaste_engine == 0:
            engine_status = "Current engine: Textual default\n"
        elif self.app.copypaste_engine == 1:
            engine_status = "Current engine: Pyperclip\n"
        elif self.app.copypaste_engine == 2:
            engine_status = "Current engine: Clipman\n"

        with VerticalScroll(classes='fullscreen'):
            with Container(classes='fullscreen container bordered_primary'):
                yield Label(clipman_status, classes='screen_label h1')
                yield Label(pyperclip_status, classes='screen_label h1')
                yield Label(self.instructions, classes='screen_label')
                yield Label(engine_status, classes='screen_label')
                yield TextArea(id="test_box", classes="link_box")
                yield Label('', id='test_label', classes='screen_label h2')
                yield Label(
                    "Note: Textual default can't be tested directly. \n Just try it out and see if it works.", 
                    classes='screen_label'
                )
                with Horizontal(classes='screen_buttonbar'):
                    yield Button("Test Clipman", id='clipman_test_button', classes='screen_button')
                    yield Button("Test Pyperclip", id='pyperclip_test_button', classes='screen_button')
                    yield Button('Close', id='close_button', classes='screen_button')

    @on(Button.Pressed, selector='#clipman_test_button')
    def run_clipman_test(self):

        test_text = self.query_one('#test_box').text
        paste_text = None
        try:
            paste_text = clipman.paste()
        except Exception as e:
            self.log.error(e)
            # Because this is a test box, we want to silence the error.

        if test_text == paste_text:
            self.query_one('#test_label').update('[green]Clipman test passed.[/green]')
            self.app.notify("Copy/Paste test successful.")
        else:
            self.query_one('#test_label').update('[red]Clipman test failed.[/red]')
            self.app.notify("Copy/Paste test failed.")

    @on(Button.Pressed, selector='#pyperclip_test_button')
    def run_pyperclip_test(self):

        test_text = self.query_one('#test_box').text
        paste_text = None
        try:
            paste_text = pyperclip.paste()
        except Exception as e:
            self.log.error(e)
            # Because this is a test box, we want to silence the error.

        if test_text == paste_text:
            self.query_one('#test_label').update('[green]Pyperclip test passed.[/green]')
            self.app.notify("Copy/Paste test successful.")
        else:
            self.query_one('#test_label').update('[red]Pyperclip test failed.[/red]')
            self.app.notify("Copy/Paste test failed.")

    @on(Button.Pressed, selector='#close_button')
    def window_close(self):
        self.dismiss()
    

class MessageScreen(TextualdonModalScreen):
    """ Generic screen used in two places. | Callbacks: None \n
    Called by:
    - `PortInput.action_info` (settings.py)
    - `WelcomeWidget.show_alpha_message` (widgets.py)"""

    BINDINGS = [
        Binding("enter", "dismiss", description="Close the pop-up screen.", show=True),
    ]
    controls = "Press enter, esc, or click anywhere to close."

    def __init__(self, message: str, **kwargs):
        super().__init__(**kwargs)
        self.message = message

    def compose(self):
        with Container(classes='screen_container wide'):
            yield Label(self.message, classes='screen_label')

    def on_click(self):
        self.dismiss()

class RoadmapScreen(TextualdonModalScreen):

    BINDINGS = [
        Binding("enter", "dismiss", description="Close the pop-up screen.", show=True),
    ]
    controls = "Press enter, esc, or click anywhere to close."

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        roadmap_path = Path(resources.files(self.app.app_name) / "Roadmap.md")
        with open(roadmap_path, 'r') as f:
            self.roadmap = f.read()

    def compose(self):
        with VerticalScroll(classes='fullscreen'):
            yield Markdown(self.roadmap, classes='fullscreen container bordered_primary')

    def on_mount(self):
        self.query_one(VerticalScroll).focus()

    def on_click(self):
        self.dismiss()




from mastodon import Mastodon

from rich.text import Text
from textual.worker import Worker
from textual.dom import DOMNode

from textualdon.error_handler import SafeModeError


class MastodonProxy(DOMNode):
    """ A proxy class for the Mastodon API wrapper. Makes the API async.
    Runs all API calls in a worker."""

    def __init__(self, mastodon_instance: Mastodon):
        self.mastodon = mastodon_instance
        print(f"Proxy class initialized for Mastodon object: {type(self.mastodon)}")
    
    def check_safe_mode(self):

        if self.app.safe_mode:
            with self.app.capture_exceptions():
                raise SafeModeError("Safe Mode is enabled.")

    async def wrapped_method(self, attribute, *args, **kwargs):

        self.check_safe_mode()
        try:
            worker:Worker = self.app.run_api_call(attribute, *args, **kwargs)
            await worker.wait()
        except Exception as e:
            self.log.error(Text(f"Error in API call: {e}", style="bold red"))
            raise e
        else:
            return worker.result
        finally:
            self.log.debug(f"wrapped_method completed with type: {type(worker.result)}")


    # NOTE: I much would have preferred to use __getattr__ here but something about it
    # just doesn't work. It seems to make Textual upset. So had to resort to manual
    # bridge methods for every API call.      
        
    async def app_verify_credentials(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'app_verify_credentials')
        return await self.wrapped_method(attribute, *args, **kwargs)

    async def account_verify_credentials(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'account_verify_credentials')
        return await self.wrapped_method(attribute, *args, **kwargs)        

    async def timeline(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'timeline')
        return await self.wrapped_method(attribute, *args, **kwargs)

    async def trending_tags(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'trending_tags')
        return await self.wrapped_method(attribute, *args, **kwargs)

    async def trending_links(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'trending_links')
        return await self.wrapped_method(attribute, *args, **kwargs)

    async def trending_statuses(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'trending_statuses')
        return await self.wrapped_method(attribute, *args, **kwargs)

    async def bookmarks(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'bookmarks')
        return await self.wrapped_method(attribute, *args, **kwargs)
    
    async def favorites(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'favorites')
        return await self.wrapped_method(attribute, *args, **kwargs)
    
    async def status(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'status')
        return await self.wrapped_method(attribute, *args, **kwargs)

    async def status_context(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'status_context')
        return await self.wrapped_method(attribute, *args, **kwargs)

    async def status_update(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'status_update')
        return await self.wrapped_method(attribute, *args, **kwargs)
    
    async def status_reply(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'status_reply')
        return await self.wrapped_method(attribute, *args, **kwargs)
    
    async def status_post(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'status_post')
        return await self.wrapped_method(attribute, *args, **kwargs)
    
    async def log_in(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'log_in')
        return await self.wrapped_method(attribute, *args, **kwargs)

    async def auth_request_url(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'auth_request_url')
        return await self.wrapped_method(attribute, *args, **kwargs)

    async def account_relationships(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'account_relationships')
        return await self.wrapped_method(attribute, *args, **kwargs)

    async def revoke_access_token(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'revoke_access_token')
        return await self.wrapped_method(attribute, *args, **kwargs)
    
    async def account_follow(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'account_follow')
        return await self.wrapped_method(attribute, *args, **kwargs)

    async def account_unfollow(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'account_unfollow')
        return await self.wrapped_method(attribute, *args, **kwargs)
    
    async def account_block(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'account_block')
        return await self.wrapped_method(attribute, *args, **kwargs)
    
    async def account_unblock(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'account_unblock')
        return await self.wrapped_method(attribute, *args, **kwargs)
    
    async def account_mute(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'account_mute')
        return await self.wrapped_method(attribute, *args, **kwargs)
    
    async def account_unmute(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'account_unmute')
        return await self.wrapped_method(attribute, *args, **kwargs)
    
    async def status_pin(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'status_pin')
        return await self.wrapped_method(attribute, *args, **kwargs)
    
    async def status_unpin(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'status_unpin')
        return await self.wrapped_method(attribute, *args, **kwargs)
    
    async def status_mute(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'status_mute')
        return await self.wrapped_method(attribute, *args, **kwargs)
    
    async def status_unmute(self, *args, **kwargs):    
        attribute = getattr(self.mastodon, 'status_unmute')
        return await self.wrapped_method(attribute, *args, **kwargs)
    
    async def status_delete(self, *args, **kwargs):    
        attribute = getattr(self.mastodon, 'status_delete')
        return await self.wrapped_method(attribute, *args, **kwargs)
    
    async def status_unblog(self, *args, **kwargs):    
        attribute = getattr(self.mastodon, 'status_unblog')
        return await self.wrapped_method(attribute, *args, **kwargs)
    
    async def status_reblog(self, *args, **kwargs):    
        attribute = getattr(self.mastodon, 'status_reblog')
        return await self.wrapped_method(attribute, *args, **kwargs)
    
    async def status_favourite(self, *args, **kwargs):
        attribute = getattr(self.mastodon, 'status_favourite')
        return await self.wrapped_method(attribute, *args, **kwargs)

    async def status_unfavourite(self, *args, **kwargs):    
        attribute = getattr(self.mastodon, 'status_unfavourite')
        return await self.wrapped_method(attribute, *args, **kwargs)
    
    async def status_bookmark(self, *args, **kwargs):    
        attribute = getattr(self.mastodon, 'status_bookmark')
        return await self.wrapped_method(attribute, *args, **kwargs)
    
    async def status_unbookmark(self, *args, **kwargs):    
        attribute = getattr(self.mastodon, 'status_unbookmark')
        return await self.wrapped_method(attribute, *args, **kwargs)





# Standard Library imports
from typing import cast
import random
from dataclasses import dataclass

# Third Party imports
from mastodon import Mastodon
# from textual_spinbox import SpinBox

# Textual imports
from textual import on
from textual.messages import Message
from textual.binding import Binding
from textual.containers import Container, Horizontal
from textual.widgets import Static, Switch, Select, Label, Input
from textual.widget import Widget
from textual.validation import Number, ValidationResult

# TextualDon imports
from textualdon.messages import (
    SuperNotify,
    LoginStatus,
    EnableSafeMode,
    TriggerRandomError,
    SwitchMainContent,
    DeleteLogs
)
from textualdon.sql import SQLite
from textualdon.simplebutton import SimpleButton
from textualdon.screens import CopyPasteTester, MessageScreen
# from textualdon.widgets import InputCustom


class PortInput(Input):

    BINDINGS = [Binding("ctrl+k", "info", "Toggle info")]

    port_info = """
TextualDon will choose a random port between 49152 and 65535 for the callback port \
when it first launches. If this creates a conflict for you, you can change the port \
here. The port must be a number between 1024 and 65535.
"""

    @dataclass
    class Blur(Message):
        """Posted when the input is blurred.

        Can be handled using `on_input_blur` in a subclass of `Input` or in a
        parent widget in the DOM.
        """

        input: Input
        """The `Input` widget that is being submitted."""
        value: str
        """The value of the `Input` being submitted."""
        validation_result: ValidationResult | None = None
        """The result of validating the value on submission, formed by combining the results for each validator.
        This value will be None if no validation was performed, which will be the case if no validators are supplied
        to the corresponding `Input` widget."""

        @property
        def control(self) -> Input:
            """Alias for self.input."""
            return self.input

    def _on_blur(self):

        self._pause_blink()

        validation_result = (
            self.validate(self.value) if "blur" in self.validate_on else None
        )
        self.post_message(self.Blur(self, self.value, validation_result))    

    def action_info(self) -> None:
        self.app.push_screen(MessageScreen(self.port_info, classes="modal_screen"))


class Settings(Widget):
    """This is a simple widget that displays the login settings for the user.
    It's used in the LoginPage class to display the login settings."""

    class ChangeHatching(Message):
        """Message to change the hatching of the main content."""
        def __init__(self, event: Select.Changed) -> None:
            super().__init__()
            self.changed = event

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.mastodon = cast(Mastodon, self.app.mastodon)
        self.db = cast(SQLite, self.app.sqlite)

        sql_query = "SELECT value FROM settings WHERE name = ?"

        row1 = self.db.fetchone(sql_query, ("first_launch",))
        row2 = self.db.fetchone(sql_query, ("auto_login",))
        row3 = self.db.fetchone(sql_query, ("auto_load",))
        row4 = self.db.fetchone(sql_query, ("show_images",))
        row5 = self.db.fetchone(sql_query, ("link_behavior",))
        row6 = self.db.fetchone(sql_query, ("copypaste_engine",))
        row7 = self.db.fetchone(sql_query, ("show_on_startup",))
        row8 = self.db.fetchone(sql_query, ("hatching",))
        row9 = self.db.fetchone(sql_query, ("callback_port",))

        first_launch: bool = ('True' == row1[0])
        self.auto_login:   bool = ('True' == row2[0])
        self.auto_load:    bool = ('True' == row3[0])
        self.show_images:  bool = ('True' == row4[0])
        self.link_behavior:    int = int(row5[0])  # 0 = browser, 1 = clipboard, 2 = manual
        self.copypaste_engine: int = int(row6[0])  # 0 = textual default, 1 = pyperclip, 2 = clipman 
        self.show_on_startup: str = row7[0]
        self.hatching:        str = row8[0]
        self.callback_port:   str = row9[0]   # this is a number, but Input expects a string

        if first_launch:
            self.db.update_column("settings", "value", "False", "name", "first_launch")
            self.callback_port = str(random.randint(49152, 65535))
            self.db.update_column("settings", "value", self.callback_port, "name", "callback_port")
        
        self.log.debug(
            f"first_launch: {first_launch} \n"
            f"auto_login: {self.auto_login} \n"
            f"auto_load: {self.auto_load} \n"
            f"show_images: {self.show_images} \n"
            f"link_behavior: {self.link_behavior} \n"
            f"show_on_startup: {self.show_on_startup} \n"
            f"hatching: {self.hatching} \n"
            f"callback_port: {self.callback_port} \n"
        )

        #~ Options lists ~#
        def format(x):
            return x.replace("_", " ").title()

        select_page_options = [
            "login_page", "home", "notifications", "explore", "live_feeds",
            "private_mentions", "bookmarks", "favorites", "lists"
        ]
        self.select_page_options = [(format(option), option) for option in select_page_options]

        hatching_options = ["none", "left", "right", "cross"]
        self.hatching_options = [(format(option), option) for option in hatching_options]

        self.link_options = [
            ("Open in browser", 0),
            ("Copy to clipboard", 1),
            ("Manual", 2)]
        
        self.copypaste_options = [
            ("Textual default", 0),
            ("Pyperclip", 1),
            ("Clipman", 2)]

        self.border_title = "Settings"

    def compose(self):

        yield Static("[middle]Settings with * have a help pop-up toggled with ctrl-k", classes="short_label")
        with Container(classes="settings_container"):
            yield Static("Sign out of current account", classes="settings_text")
            yield SimpleButton(" Logout ", id="logout", classes="settings_button bordered")

            yield Static("\nSign in automatically. \n (Uses most recent sign-in)", classes="settings_text")
            yield Switch(self.auto_login, id="auto_login", classes="settings_button")

            yield Static("Auto-load page content", classes="settings_text")
            yield Switch(self.auto_load, id="auto_load", classes="settings_button")

            yield Static("Show page on startup", classes="settings_text")
            yield Select(
                self.select_page_options,
                value=self.show_on_startup,
                id="show_on_startup",
                classes="settings_list",
                allow_blank=False
            ) 

            yield Static("Download/Show images", classes="settings_text")
            yield Switch(self.show_images, id="show_images", classes="settings_button")

            yield Static("Link behavior", id="clicklink_desc", classes="settings_text")
            yield Select(
                self.link_options,
                value=self.link_behavior,
                id="link_behavior",
                classes="settings_list",
                allow_blank=False
            )

            yield Static("Copy/Paste engine", classes="settings_text")
            yield Select(
                self.copypaste_options,
                value=self.copypaste_engine,
                id="copypaste_engine",
                classes="settings_list",
                allow_blank=False
            )

            yield Static("Copy/Paste Tester", id="copytest_label", classes="settings_text")
            yield SimpleButton(" Open ", id="copy_paste_tester", classes="settings_button bordered")

            yield Static("Background hatching", classes="settings_text")
            yield Select(
                self.hatching_options,
                value=self.hatching,
                id="hatching",
                classes="settings_list",
                allow_blank=False
            )

            yield Static("Callback Port* \nMust be number between 1024 - 65535", classes="settings_text")
            yield PortInput(
                value=self.callback_port,
                type="integer",
                max_length=5,
                validators=Number(minimum=1024, maximum=65535),
                validate_on=["blur"]
            )

            yield Static("Reset all startup warnings", classes="settings_text")    
            yield SimpleButton(" Reset ", id="reset_warnings", classes="settings_button bordered")

            yield Static("View Development settings", classes="settings_text")    
            yield SimpleButton(" View ", id="view_dev", classes="settings_button bordered")

            # Remember to make the grid bigger with each additional setting
            # settings.tcss / #settings_container - increment grid rows by one for each new setting
            # also must adjust height of Settings class manually  (auto mode not working)

    def on_mount(self):

        issues = 0
        if not self.app.clipman_works:
            issues += 1
        if not self.app.pyperclip_works:
            issues += 1

        if issues == 0:
            self.query_one("#copytest_label").update(
                "Copy/Paste Tester \n Test Status: [green](All Passed)[/green]"
            )
        elif issues == 1:
            self.query_one("#copytest_label").update(
                "Copy/Paste Tester \n Test Status: [yellow](1 Issue)[/yellow]"
            )
        elif issues == 2:
            self.query_one("#copytest_label").update(
                "Copy/Paste Tester \n Test Status: [red](2 Issues)[/red]"
            )

    @on(SimpleButton.Pressed, "#logout")
    def logout(self) -> None:

        self.post_message(SuperNotify("Logged out."))
        self.app.mastodon = None
        self.app.logged_in_user_id = None
        self.post_message(LoginStatus(
            statusbar="Status: Offline",
            loginpage_message="Logged out.",
        ))

    @on(Switch.Changed, "#auto_login")
    def auto_login(self, event: Switch.Changed) -> None:

        self.db.update_column("settings", "value", str(event.value), "name", "auto_login")
        self.notify(f"Auto-login set to {event.value}")

    @on(Switch.Changed, "#auto_load")
    def auto_load(self, event: Switch.Changed) -> None:

        self.db.update_column("settings", "value", str(event.value), "name", "auto_load")
        self.app.autoload_value = event.value
        self.notify(f"Auto-load set to {event.value}")

    @on(Select.Changed, "#show_on_startup")
    def show_on_startup(self, event: Select.Changed) -> None:

        self.db.update_column("settings", "value", str(event.value), "name", "show_on_startup")

    @on(Switch.Changed, "#show_images")
    def show_images(self, event: Switch.Changed) -> None:
        
        self.db.update_column("settings", "value", str(event.value), "name", "show_images")
        self.app.show_images = event.value
        self.notify(f"Show images set to {event.value}")

    @on(Select.Changed, "#link_behavior")
    def change_link_behavior(self, event: Select.Changed) -> None:
        
        self.db.update_column("settings", "value", str(event.value), "name", "link_behavior")
        self.app.link_behavior = event.value
        self.log.debug(f"self.app.link_behavior: {self.app.link_behavior}")
        desc_static = cast(Label, self.query_one("#clicklink_desc"))
        if event.value == 0:
            desc_static.update("Link behavior: \nOpen URL in a new browser tab.")
        elif event.value == 1:
            desc_static.update("Link behavior: \nCopy link to clipboard.")
        elif event.value == 2:
            desc_static.update("Link behavior: \nProvides pop-up to copy link manually.")
        desc_static.refresh()

    @on(Select.Changed, "#copypaste_engine")
    def copy_paste_engine(self, event: Select.Changed) -> None:
        
        self.db.update_column("settings", "value", str(event.value), "name", "copypaste_engine")
        self.app.copypaste_engine = event.value
        self.log.debug(f"self.app.copypaste_engine: {self.app.copypaste_engine}")


    @on(SimpleButton.Pressed, "#copy_paste_tester")
    def open_tester_screen(self) -> None:
        self.app.push_screen(CopyPasteTester(classes="fullscreen"))

    @on(Select.Changed, "#hatching")
    def change_hatching(self, event: Select.Changed) -> None:

        self.db.update_column("settings", "value", str(event.value), "name", "hatching")
        self.post_message(self.ChangeHatching(event))

    @on(PortInput.Blur)
    def update_port(self, event: PortInput.Blur) -> None:

        if event.value != self.callback_port:

            if event.validation_result.is_valid:
                self.log.debug(f"event.value type: {type(event.value)}")        #! TODO remove me.
                self.db.update_column("settings", "value", str(event.value), "name", "callback_port")
                self.notify(f"Callback port set to {event.value}")
                self.callback_port = event.value
            else:
                self.notify(str(event.validation_result.failure_descriptions[0]))
                event.input.value = self.callback_port
        
    @on(SimpleButton.Pressed, "#reset_warnings")
    def show_warnings(self) -> None:

        self.db.update_column("settings", "value", "False", "name", "warning_checkbox_wsl")
        self.db.update_column("settings", "value", "False", "name", "warning_checkbox_first")
        self.notify("Startup Warnings re-enabled.")

    @on(SimpleButton.Pressed, "#view_dev")
    def view_dev_settings(self) -> None:
        self.post_message(SwitchMainContent("dev_settings"))

class DevSettings(Widget):
    """This is a simple widget that displays the developer settings for the user.
    It's used in the LoginPage class to display the developer settings."""

    def compose(self):

        self.db = cast(SQLite, self.app.sqlite)

        sql_query = "SELECT value FROM settings WHERE name = ?"
        row1 = self.db.fetchone(sql_query, ("view_json_active",))
        view_json_active: bool = ('True' == row1[0])
            
        self.border_title = "Developer Settings"

        with Container(classes="settings_container dev"):
            yield Static("Trigger a random exception", classes="settings_text")
            yield SimpleButton(" Bring it ", id="test_error", classes="settings_button bordered")

            yield Static("Enable safe mode", classes="settings_text")
            yield SimpleButton(" Enable ", id="test_safemode", classes="settings_button bordered")

            yield Static("Show button to view JSON \nin toot options menu", classes="settings_text")
            yield Switch(view_json_active, id="view_json_active", classes="settings_button")

            yield Static("Delete all log files in logs folder", classes="settings_text")
            yield SimpleButton( " Delete ", id="delete_logs", classes="settings_button bordered")

            # Remember to make the grid bigger with each additional setting
            # settings.tcss / #settings_container - increment grid rows by one for each new setting
            # also must adjust height of Settings class manually  (auto mode not working)

    @on(SimpleButton.Pressed, "#test_error")
    def trigger_mock_error(self):
        self.post_message(TriggerRandomError())

    @on(SimpleButton.Pressed, "#test_safemode")
    def trigger_safe_mode(self):
        self.post_message(EnableSafeMode())

    @on(Switch.Changed, "#view_json_active")
    def auto_login(self, event: Switch.Changed) -> None:

        self.db.update_column("settings", "value", str(event.value), "name", "view_json_active")
        if event.value:
            self.notify("View JSON button enabled.")

    @on(SimpleButton.Pressed, "#delete_logs")
    async def delete_logs(self) -> None:

        self.post_message(DeleteLogs())


from textual.message import Message
from textual.widget import Widget

class UpdateBannerMessage(Message):
    """Message to update the banner with a new message.
    ```
    from textualdon.messages import UpdateBannerMessage
    self.post_message(UpdateBannerMessage(f"Hello, {user}!"))
    ``` """
    def __init__(self, message: str) -> None:
        super().__init__()
        self.message = str(message)

class SuperNotify(Message):
    """Triggers a notification and updates the banner at the same time."""
    def __init__(self, message: str) -> None:
        super().__init__()
        self.message = str(message)

class LoginStatus(Message):
    """Message to update the login status in the Oauth widget.
    ```
    from textualdon.messages import LoginStatus
    self.post_message(LoginStatus("Logged in successfully"))
    ``` """
    def __init__(
        self, 
        statusbar: str,
        loginpage_message: str,
        instance_url: str = None
    ) -> None:
        super().__init__()
        self.statusbar = statusbar
        self.loginpage_message = loginpage_message
        self.instance_url = instance_url

class RefreshCurrentPage(Message):
    """Message to refresh the current page.
    ```
    from textualdon.messages import RefreshCurrentPage
    self.post_message(RefreshCurrentPage())
    ``` """
    pass

class SwitchMainContent(Message):
    """Message to switch the main content.
    ```
    from textualdon.messages import SwitchMainContent
    self.post_message(SwitchMainContent('home'))
    ``` """
    def __init__(self, content: str) -> None:
        super().__init__()
        self.content = content

class ExamineToot(Message):
    """Message to examine a toot.
    ```
    from textualdon.messages import ExamineToot
    self.post_message(ExamineToot(toot_id))
    ``` """
    def __init__(self, toot_id: int) -> None:
        super().__init__()
        self.toot_id = toot_id

class LoginComplete(Message):
    """Message to signal that the login process is complete."""
    pass

class UserPopupMessage(Message):
    """ ``` \n
    message must be either 'follow' or 'profile'

    from textualdon.messages import UserPopupMessage
    self.post_message(UserPopupMessage('follow', self.account, self.relation))
    ``` """
    def __init__(self, message: str, account_dict: dict, relation_dict: dict) -> None:
        super().__init__()
        self.message = message
        self.account = account_dict
        self.relation = relation_dict

class OpenCallbackScreen(Message):
    def __init__(self, auth_url: str) -> None:
        self.auth_url = auth_url
        super().__init__()

class CallbackSuccess(Message):
    """ ``` \n
    from textualdon.messages import CallbackSuccess
    self.post_message(CallbackSuccess())
    ``` """
    pass

class CallbackCancel(Message):
    """ ``` \n
    from textualdon.messages import CallbackCancel
    self.post_message(CallbackCancel())
    ``` """
    pass

class ScrollToWidget(Message):
    def __init__(self, widget: Widget) -> None:
        super().__init__()
        self.widget = widget

class EnableSafeMode(Message):
    pass

class TriggerRandomError(Message):
    pass

class ExceptionMessage(Message):
    def __init__(self, exception: Exception) -> None:
        super().__init__()
        self.exception = exception

class DeleteLogs(Message):
    pass

class OpenRoadmap(Message):
    pass